# Copyright 2023 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from urllib.parse import urlparse

import clickhouse_driver

from . import consts


def get_clickhouse_url():
    ch_url = os.environ.get('KRAKEN_CLICKHOUSE_URL', consts.DEFAULT_CLICKHOUSE_URL)
    return ch_url


def get_clickhouse():
    ch_url = get_clickhouse_url()
    o = urlparse(ch_url)
    ch = clickhouse_driver.Client(host=o.hostname)
    return ch



# Copyright 2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import shutil
import logging
import zipfile
import tempfile
import subprocess

log = logging.getLogger(__name__)


def run(cmd, check=True, cwd=None, capture_output=False, text=None):
    log.info("execute '%s' in '%s'", cmd, cwd)
    p = subprocess.run(cmd, shell=True, check=check, cwd=cwd, capture_output=capture_output, text=text)
    return p


def install_reqs(tool_file):
    tool_dir = os.path.abspath(os.path.dirname(tool_file))
    vendor_dir = os.path.join(tool_dir, 'vendor')
    reqs_txt_path = os.path.join(tool_dir, 'requirements.txt')

    # install requirements to vendor dir
    if os.path.exists(reqs_txt_path):
        if os.path.exists(vendor_dir):
            shutil.rmtree(vendor_dir)
        os.makedirs(vendor_dir)

        cmd = 'python3 -m pip install -r requirements.txt --target=./vendor'
        run(cmd, cwd=tool_dir)

    return tool_dir


def package_tool(tool_file):
    # load file and parse as JSON
    with open(tool_file) as fp:
        meta = json.load(fp)

    tool_dir = install_reqs(tool_file)

    # store tool files in zip package
    tf = tempfile.NamedTemporaryFile(prefix='kkci-pkg-', suffix='.zip', delete=True)
    with zipfile.ZipFile(tf, "w") as pz:
        for root, _, files in os.walk(tool_dir):
            for name in files:
                if name.endswith(('.pyc', '~')):
                    continue
                if name == 'tool.zip':
                    continue
                p = os.path.join(root, name)
                n = os.path.relpath(p, tool_dir)
                pz.write(p, arcname=n)

    tf.seek(0)

    return meta, tf, len(pz.namelist())



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import logging

# AWS
import boto3

from .aws import login_to_aws
from .common import _create_agent


log = logging.getLogger(__name__)


# AWS ECS FARGATE #############################################################

def create_fargate_tasks(ag, system, num,
                         server_url, clickhouse_addr):
    credential = login_to_aws()
    if not credential:
        return

    _, aws = ag.get_deployment()

    region = aws['region']
    ec2 = boto3.client("ec2", region_name=region, **credential)
    ecs = boto3.client('ecs', region_name=region, **credential)

    system_norm = system.name.replace(':', '_').replace('/', '_').replace('.', '_')
    task_def_name = 'kraken-agent-1-%s' % system_norm

    # if there is no task definition yet then create one
    response = ecs.list_task_definitions(familyPrefix=task_def_name)
    if len(response['taskDefinitionArns']) == 0:
        ecs.register_task_definition(
            family=task_def_name,
            containerDefinitions=[{
                "name": "kraken-agent",
                "image": system.name,
                "logConfiguration": {
                    "logDriver": "awslogs",
                    "options": {
                        "awslogs-group": "/ecs/kraken-agent-definition",
                        "awslogs-region": region,
                        "awslogs-stream-prefix": "ecs"
                    }
                },
            }],
            requiresCompatibilities=['FARGATE'],
            memory="512",
            cpu="256",
            networkMode="awsvpc",
            executionRoleArn="arn:aws:iam::214272040963:role/ecsTaskExecutionRole",
        )

    # prepare command for container
    cmd = 'apt-get update && apt-get install -y --no-install-recommends ca-certificates sudo wget python3'
    cmd += ' && mkdir -p /opt/kraken'
    cmd += ' && wget -O /opt/kraken/kkagent {server_url}/bk/install/agent'
    cmd += ' && wget -O /opt/kraken/kktool {server_url}/bk/install/tool'
    cmd += ' && chmod a+x /opt/kraken/kkagent /opt/kraken/kktool'
    cmd += ' && mkdir -p /tmp/kk-jobs'
    cmd += ' && /opt/kraken/kkagent run -d /tmp/kk-jobs -s {server_url} -c {clickhouse_addr}'
    cmd += ' --system-id {system_id} --one-job'
    cmd = cmd.format(server_url=server_url, clickhouse_addr=clickhouse_addr,
                     system_id=system.id)

    cluster = aws['cluster']
    subnets = aws['subnets'].split(',')
    security_groups = aws['security_groups'].split(',')

    response = ecs.run_task(
        taskDefinition=task_def_name,
        count=num,
        cluster=cluster,
        launchType="FARGATE",
        networkConfiguration={
            'awsvpcConfiguration': {
                'subnets': subnets,
                'securityGroups': security_groups,
                'assignPublicIp': 'ENABLED'
            }
        },
        overrides={
            'containerOverrides': [{
                'name': 'kraken-agent',
                'command': [
                    'bash', '-c', cmd
                ],
                'environment': [
                    {'name': 'KRAKEN_SERVER_ADDR', 'value': server_url},
                    {'name': 'KRAKEN_CLICKHOUSE_ADDR', 'value': clickhouse_addr},
                ],
                #'cpu': 123,
                #'memory': 123,
                #'memoryReservation': 123,
            }],
            #'cpu': 'string',
            #'memory': 'string',
        }
    )

    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        log.warning('some problem occured')
        log.warning(response)
        return

    all_tasks = [t['taskArn'] for t in response['tasks']]
    tasks_ready = []
    tasks_not_ready = all_tasks[:]
    erred_task = None
    while len(tasks_not_ready) > 0 and not erred_task:
        resp = ecs.describe_tasks(cluster=cluster,
                                  tasks=tasks_not_ready)
        tasks_ready = []
        tasks_not_ready = []
        for task in resp['tasks']:
            if task['lastStatus'] == 'STOPPED':
                erred_task = task
                break
            if task['lastStatus'] != 'RUNNING':
                tasks_not_ready.append(task['taskArn'])
                continue
            eni = None
            for field in task['attachments'][0]['details']:
                if field['name'] == 'networkInterfaceId':
                    eni = field['value']
            if not eni:
                tasks_not_ready.append(task['taskArn'])
                continue
            iface_resp = ec2.describe_network_interfaces(NetworkInterfaceIds=[eni])
            priv_ip = iface_resp['NetworkInterfaces'][0]['PrivateIpAddress']
            pub_dns = iface_resp['NetworkInterfaces'][0]['Association']['PublicDnsName']
            pub_ip = iface_resp['NetworkInterfaces'][0]['Association']['PublicIp']
            name = '.'.join(pub_dns.split('.')[:2])
            tasks_ready.append(dict(task_arn=task['taskArn'],
                                    name=name,
                                    address=priv_ip,
                                    ip_address=pub_ip))
        if tasks_not_ready:
            time.sleep(2)

    if erred_task:
        msg = erred_task['containers'][0]['reason']
        log.warning('problem with starting task %s: %s', erred_task['taskArn'], msg)
        # stop all tasks
        for t in all_tasks:
            try:
                ecs.stop_task(cluster=cluster,
                              task=t,
                              reason='stopping other tasks due to an error')
            except Exception:
                log.exception('IGNORED EXCEPTION')

    for task in tasks_ready:
        address = task['address']
        params = dict(name=task['name'],
                      address=address,
                      ip_address=task['ip_address'],
                      extra_attrs=dict(system=0, task_arn=task['task_arn']))
        a = _create_agent(params, ag)
        log.info('spawned new agent %s on ECS Fargate task %s', a, task)


def destroy_fargate_task(ag, agent):  # pylint: disable=unused-argument
    credential = login_to_aws()
    if not credential:
        return

    _, depl = ag.get_deployment()

    region = depl['region']
    ecs = boto3.client('ecs', region_name=region, **credential)

    task_arn = agent.extra_attrs['task_arn']
    log.info('terminate ecs task %s', task_arn)
    try:
        ecs.stop_task(cluster=depl['cluster'],
                      task=task_arn,
                      reason='stopping task that completed the job')
    except Exception:
        log.exception('IGNORED EXCEPTION')



"""added relevancy to test case result

Revision ID: 770cfcc80cfb
Revises: d524cbcbf19c
Create Date: 2021-03-06 08:50:49.401290

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '770cfcc80cfb'
down_revision = 'd524cbcbf19c'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('test_case_results', sa.Column('relevancy', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('test_case_results', 'relevancy')



"""create Secret table

Revision ID: 1cf31ab18f84
Revises: ba436d491bf8
Create Date: 2019-12-24 07:56:00.751309

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '1cf31ab18f84'
down_revision = 'ba436d491bf8'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('secrets',
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('updated', sa.DateTime(), nullable=False),
                    sa.Column('deleted', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.Unicode(length=255), nullable=True),
                    sa.Column('project_id', sa.Integer(), nullable=False),
                    sa.Column('kind', sa.Integer(), nullable=True),
                    sa.Column('data', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
                    sa.ForeignKeyConstraint(['project_id'], ['projects.id'], ),
                    sa.PrimaryKeyConstraint('id'))


def downgrade():
    op.drop_table('secrets')



"""user is session nullable

Revision ID: 2996b1b6e227
Revises: 87358b464400
Create Date: 2022-10-22 09:39:57.125878

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2996b1b6e227'
down_revision = '87358b464400'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('user_sessions', 'user_id',
               existing_type=sa.INTEGER(),
               nullable=True)


def downgrade():
    op.alter_column('user_sessions', 'user_id',
               existing_type=sa.INTEGER(),
               nullable=False)



# Copyright 2022-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import logging
import urllib.parse

# LDAP
import ldap
import ldap.filter

# OAuth & OIDC
from authlib.integrations.base_client import BaseOAuth, FrameworkIntegration, OAuth2Mixin, OpenIDMixin, BaseApp
from authlib.integrations.requests_client import OAuth2Session  # OAuth1Session,

from werkzeug.exceptions import Unauthorized
from flask import request, redirect

from .models import db, User, UserSession, get_setting, get_settings_group


log = logging.getLogger(__name__)


def _get_ldap_settings():
    idp_settings = get_settings_group('idp')
    return (idp_settings['ldap_enabled'],
            idp_settings['ldap_server'],
            idp_settings['bind_dn'],
            idp_settings['bind_password'],
            idp_settings['base_dn'],
            idp_settings['search_filter'])


def authenticate_ldap(username, password):
    enabled, ldap_server, bind_dn, bind_pswd, base_dn, search_filter = _get_ldap_settings()
    if not enabled:
        return None
    try:
        conn = ldap.initialize(ldap_server)
        conn.set_option(ldap.OPT_REFERRALS, 0)
        conn.simple_bind_s(bind_dn, bind_pswd)
    except ldap.INVALID_CREDENTIALS as ex:
        conn.unbind_s()
        raise Exception('LDAP: incorrect bind credentials') from ex
    except ldap.SERVER_DOWN as ex:
        raise Exception('LDAP: server is down') from ex

    ldap_filter = search_filter % (ldap.filter.escape_filter_chars(username))
    attrs = ['mail']
    result = conn.search_s(base_dn, ldap.SCOPE_SUBTREE, ldap_filter, attrs)
    if not result:
        conn.unbind_s()
        log.info('ldap: user %s not found or search filter is not ok', username)
        return None

    user = result[0]
    conn.unbind_s()

    conn = ldap.initialize(ldap_server)
    conn.set_option(ldap.OPT_REFERRALS, 0)
    try:
        conn.simple_bind_s(user[0], password)
    except ldap.INVALID_CREDENTIALS:
        log.info('ldap: wrong creds for user %s (%s)', username, user[0])
        return None
    finally:
        conn.unbind_s()

    email_b = user[1].get('mail', [b''])
    email = email_b[0].decode('utf-8')

    return dict(username=username, email=email)


def check_ldap_settings():
    enabled, ldap_server, bind_dn, bind_pswd, _, _ = _get_ldap_settings()
    if not enabled:
        return 'ok'
    try:
        conn = ldap.initialize(ldap_server)
        conn.set_option(ldap.OPT_REFERRALS, 0)
        conn.simple_bind_s(bind_dn, bind_pswd)
    except Exception as ex:
        try:
            conn.unbind_s()
        except Exception:
            pass
        return str(ex)
    return 'ok'


def test_ldap():
    # https://www.forumsys.com/2022/05/10/online-ldap-test-server/
    # ldap.forumsys.com
    resp = authenticate_ldap('gauss', 'password')
                             # 'ldap://ldap.forumsys.com',
                             # 'cn=read-only-admin,dc=example,dc=com',
                             # 'password',
                             # 'dc=example,dc=com',
                             # 'uid=%s')
    print(resp)

    # https://github.com/rroemhild/docker-test-openldap
    # docker run --rm -p 10389:10389 -p 10636:10636 rroemhild/test-openldap
    resp = authenticate_ldap('bender', 'bender')
                             # 'ldap://localhost:10389',
                             # 'cn=admin,dc=planetexpress,dc=com',
                             # 'GoodNewsEveryone',
                             # 'ou=people,dc=planetexpress,dc=com',
                             # '(&(uid=%s)(objectClass=inetOrgPerson))')
    print(resp)

    # https://www.zflexldapadministrator.com/index.php/blog/82-free-online-ldap
    # www.zflexldap.com
    resp = authenticate_ldap('guest1', 'guest1password')
                             # 'ldap://www.zflexldap.com',
                             # 'cn=ro_admin,ou=sysadmins,dc=zflexsoftware,dc=com',
                             # 'zflexpass',
                             # 'ou=guests,dc=zflexsoftware,dc=com',
                             # '(uid=%s)')
    print(resp)


class KrakenOAuth2App(OAuth2Mixin, OpenIDMixin, BaseApp):
    client_cls = OAuth2Session


class KrakenOAuthIntegration(FrameworkIntegration):
    @staticmethod
    def load_config(oauth, name, params):
        return {}


class KrakenOAuth(BaseOAuth):
    # oauth1_client_cls = FlaskOAuth1App
    oauth2_client_cls = KrakenOAuth2App
    framework_integration_cls = KrakenOAuthIntegration


def setup_oauth(id_provider):
    idp_settings = get_settings_group('idp')

    oauth = KrakenOAuth()
    if id_provider == 'google':
        if not idp_settings['google_enabled']:
            raise Exception('Google OIDC/OAuth not enabled')
        idp = oauth.register(name='google',
                             client_id=idp_settings['google_client_id'],
                             client_secret=idp_settings['google_client_secret'],
                             server_metadata_url='https://accounts.google.com/.well-known/openid-configuration',
                             client_kwargs={'scope': 'openid email profile'})

    elif id_provider == 'microsoft':
        if not idp_settings['microsoft_enabled']:
            raise Exception('Microsoft OIDC/OAuth not enabled')
        idp = oauth.register(name='microsoft',
                             client_id=idp_settings['microsoft_client_id'],
                             client_secret=idp_settings['microsoft_client_secret'],
                             api_base_url='https://graph.microsoft.com/',
                             # server_metadata_url='https://login.microsoftonline.com/common/v2.0/.well-known/openid-configuration',
                             server_metadata_url='https://login.microsoftonline.com/organizations/v2.0/.well-known/openid-configuration',
                             client_kwargs={'scope': 'openid email profile'})

    elif id_provider == 'github':
        if not idp_settings['github_enabled']:
            raise Exception('GitHub OIDC/OAuth not enabled')
        idp = oauth.register(name='github',
                             client_id=idp_settings['github_client_id'],
                             client_secret=idp_settings['github_client_secret'],
                             api_base_url='https://api.github.com/',
                             access_token_url='https://github.com/login/oauth/access_token',
                             authorize_url='https://github.com/login/oauth/authorize',
                             client_kwargs={'scope': 'user:email'},
                             userinfo_endpoint='https://api.github.com/user')

    elif id_provider == 'auth0':
        if not idp_settings['auth0_enabled']:
            raise Exception('Auth0 OIDC/OAuth not enabled')
        idp = oauth.register(name='auth0',
                             client_id=idp_settings['auth0_client_id'],
                             client_secret=idp_settings['auth0_client_secret'],
                             client_kwargs={'scope': 'openid profile email'},
                             server_metadata_url=idp_settings['auth0_openid_config_url'])

    else:
        raise Exception('unsupported identity provider: %s' % id_provider)
    return idp


def authenticate_oidc(id_provider):
    idp = setup_oauth(id_provider)

    server_url = get_setting('general', 'server_url')
    redir_url = urllib.parse.urljoin(server_url, 'bk/oidc-logged')

    state_data = idp.create_authorization_url(redir_url)
    state_data['redirect_uri'] = redir_url
    url = state_data['url']

    return url, state_data


def oidc_logged():
    state = request.args['state']

    us = UserSession.query.filter_by(token=state, deleted=None).one_or_none()
    if us is None:
        raise Unauthorized

    idp = setup_oauth(us.details['idp'])

    if request.method == 'GET':
        error = request.args.get('error')
        if error:
            description = request.args.get('error_description')
            raise Exception("%s: %s" % (error, description))
        params = {
            'code': request.args['code'],
            'state': request.args.get('state'),
        }
    else:
        params = {
            'code': request.form['code'],
            'state': request.form.get('state'),
        }

    kwargs = {}
    claims_options = kwargs.pop('claims_options', None)
    state_data = us.details['state']
    params = idp._format_state_params(state_data, params)
    token = idp.fetch_access_token(**params, **kwargs)

    if 'id_token' in token and 'nonce' in state_data:
        userinfo = idp.parse_id_token(token, nonce=state_data['nonce'], claims_options=claims_options)
    else:
        userinfo = idp.userinfo(token=token)
    token['userinfo'] = userinfo

    user_email = token['userinfo']['email']

    user = User.query.filter_by(name=user_email).one_or_none()
    if not user:
        details = dict(idp=us.details['idp'], email=user_email)
        user = User(name=user_email, password='', details=details)

    us.user = user
    us.token = uuid.uuid4().hex
    db.session.commit()

    resp = redirect('/logged')
    resp.set_cookie('kk_session_token', us.token)

    return resp


def oidc_logout(session_details):
    idp = setup_oauth(session_details['idp'])
    logout_uri = idp.load_server_metadata().get('end_session_endpoint')
    if logout_uri:
        #return_url = url_join(request.url_root, return_url)
        #query = url_encode({'post_logout_redirect_uri': return_url})
        return redirect(logout_uri) # + '?' + query)
    return None



def test_oidc():
    authenticate_oidc('microsoft')


if __name__ == '__main__':
    # test_ldap()
    test_oidc()



"""added age to Issue

Revision ID: 0ad7507aa6d0
Revises: 98af95844f27
Create Date: 2020-01-05 13:30:13.930171

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '0ad7507aa6d0'
down_revision = '98af95844f27'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('issues', sa.Column('age', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('issues', 'age')



#!/usr/bin/env python3

# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from urllib.parse import urlparse
from xmlrpc.server import SimpleXMLRPCServer

from flask import Flask
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.cron import CronTrigger

from . import consts
from . import logs
from . import srvcheck
from . import models
from .. import version

log = logging.getLogger('planner')


class Planner:
    def __init__(self, db_url):
        job_defaults = dict(misfire_grace_time=180, coalesce=True, max_instances=1, next_run_time=None)
        self.scheduler = BackgroundScheduler(timezone='UTC', job_defaults=job_defaults)
        self.scheduler.add_jobstore('sqlalchemy', url=db_url + '?application_name=aps')
        self.scheduler.start()
        log.info('started planner scheduler')

        for idx, j in enumerate(self.get_jobs()):
            log.info('%d. name:%s trigger:%s func:%s args:%s kwargs:%s next:%s',
                     idx + 1, j['name'], j['trigger'], j['func'], j['args'], j['kwargs'], j['next_run_time'])

    def _job_to_dict(self, job):
        if isinstance(job.trigger, IntervalTrigger):
            trigger = 'interval'
        elif isinstance(job.trigger, DateTrigger):
            trigger = 'date'
        elif isinstance(job.trigger, CronTrigger):
            trigger = 'cron'
        else:
            trigger = 'unknown'
        next_run_time = ""
        if job.next_run_time:
            next_run_time = job.next_run_time.strftime("%Y-%m-%dT%H:%M:%SZ")
        return dict(id=job.id, name=job.name, func=job.func_ref, args=job.args, kwargs=job.kwargs, trigger=trigger, next_run_time=next_run_time)

    def add_job(self, func=None, trigger=None, args=None, kwargs=None, job_id=None, name=None, misfire_grace_time=None,  # pylint: disable=too-many-arguments
                coalesce=True, max_instances=None, next_run_time=None, replace_existing=False, trigger_args=None):
        if trigger_args is None:
            trigger_args = {}

        all_kw_args = dict(args=args, kwargs=kwargs, id=job_id, name=name, replace_existing=replace_existing)

        if misfire_grace_time is not None:
            all_kw_args['misfire_grace_time'] = misfire_grace_time
        if coalesce is not None:
            all_kw_args['coalesce'] = coalesce
        if max_instances is not None:
            all_kw_args['max_instances'] = max_instances
        if next_run_time is not None:
            all_kw_args['next_run_time'] = next_run_time

        if trigger in ['interval', 'repo_interval']:
            trigger = IntervalTrigger(**trigger_args)
        elif trigger == 'date':
            trigger = DateTrigger(**trigger_args)
        elif trigger == 'cron':
            trigger = CronTrigger(**trigger_args)
        else:
            raise Exception('unknown trigger type %s' % trigger)

        all_kw_args['trigger'] = trigger
        log.info('add_job args: %s', all_kw_args)

        try:
            job = self.scheduler.add_job(func, **all_kw_args)
            log.info('add_job job:%s', job)
        except Exception:
            log.exception('some problem')
            raise
        return self._job_to_dict(job)

    def get_jobs(self):
        try:
            jobs = []
            for j in self.scheduler.get_jobs():
                jobs.append(self._job_to_dict(j))
        except Exception:
            log.exception('some problem')
            raise
        return jobs

    def reschedule_job(self, job_id=None, trigger=None, trigger_args=None):
        if trigger_args is None:
            trigger_args = {}

        log.info('reschedule_job args: %s %s %s', job_id, trigger, trigger_args)

        try:
            job = self.scheduler.reschedule_job(job_id, trigger=trigger, **trigger_args)
            log.info('reschedule_job job:%s', job)
        except Exception:
            log.exception('some problem')
            raise
        return self._job_to_dict(job)

    def remove_job(self, job_id=None):
        log.info('remove_job arg: %s', job_id)

        try:
            self.scheduler.remove_job(job_id)
            log.info('remove_job done')
        except Exception:
            log.exception('some problem')
            # raise


def _db_setup(db_url):
    # Create  Flask app instance
    app = Flask('Kraken Planner')

    # Configure the SqlAlchemy part of the app instance
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url + '?application_name=planner'
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    models.db.init_app(app)

    # setup sentry
    with app.app_context():
        sentry_url = models.get_setting('monitoring', 'sentry_dsn')
        logs.setup_sentry(sentry_url)


def main():
    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
    # clickhouse_addr = os.environ.get('KRAKEN_CLICKHOUSE_ADDR', consts.DEFAULT_CLICKHOUSE_ADDR)

    srvcheck.check_postgresql(db_url)
    # srvcheck.check_tcp_service('clickhouse', clickhouse_addr, 9600)

    logs.setup_logging('planner')
    log.info('Kraken Planner started, version %s', version.version)

    # db setup
    _db_setup(db_url)

    # prepare planner and start it
    planner = Planner(db_url)

    o = urlparse(planner_url)
    planner_port = int(o.port)

    log.info('starting xml-rpc server for planner on port %d', planner_port)
    with SimpleXMLRPCServer(('0.0.0.0', planner_port), allow_none=True) as server:
        server.register_instance(planner)
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            print("\nKeyboard interrupt received, exiting.")


if __name__ == '__main__':
    main()



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

# AWS
import boto3

from ..models import get_setting


log = logging.getLogger(__name__)


# AWS EC2 ###################################################################

def check_aws_settings():
    access_key = get_setting('cloud', 'aws_access_key')
    if not access_key:
        return 'AWS access key is empty'
    if len(access_key) < 16:
        return 'AWS access key is too short'
    if len(access_key) > 128:
        return 'AWS access key is too long'

    secret_access_key = get_setting('cloud', 'aws_secret_access_key')
    if not secret_access_key:
        return 'AWS secret access key is empty'
    if len(secret_access_key) < 36:
        return 'AWS secret access key is too short'

    try:
        ec2 = boto3.client('ec2', region_name='us-east-1', aws_access_key_id=access_key, aws_secret_access_key=secret_access_key)
        ec2.describe_regions()
    except Exception as ex:
        return str(ex)

    return 'ok'


def login_to_aws():
    access_key = get_setting('cloud', 'aws_access_key')
    secret_access_key = get_setting('cloud', 'aws_secret_access_key')
    settings = ['access_key', 'secret_access_key']
    for s in settings:
        val = locals()[s]
        if not val:
            log.error('AWS %s is empty, please set it in global cloud settings', s)
            return None

    return dict(aws_access_key_id=access_key, aws_secret_access_key=secret_access_key)



"""added fields_masked to step

Revision ID: 5c16623f4d55
Revises: 9666bf8c1bb0
Create Date: 2022-11-05 18:50:29.258670

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '5c16623f4d55'
down_revision = '9666bf8c1bb0'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('steps', sa.Column('fields_masked', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('steps', 'fields_masked')



"""added repo_data to run

Revision ID: dae910bd9587
Revises: 2dd5cf2d12dc
Create Date: 2021-01-06 08:48:41.813706

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'dae910bd9587'
down_revision = '2dd5cf2d12dc'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('runs', sa.Column('repo_data', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('runs', 'repo_data')



# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import datetime

import pytz
import requests
from sqlalchemy.orm.attributes import flag_modified

# AWS
import boto3
import botocore

from ..models import db
from .. import utils
from .aws import login_to_aws
from .common import _create_agent


log = logging.getLogger(__name__)


def create_vms(ag, system, num,
               server_url, clickhouse_addr):
    credential = login_to_aws()
    if not credential:
        return

    _, aws = ag.get_deployment()

    region = aws['region']
    ec2 = boto3.client("ec2", region_name=region, **credential)
    ec2_res = boto3.resource('ec2', region_name=region, **credential)

    # get key pair
    if not ag.extra_attrs:
        ag.extra_attrs = {}
    if 'aws' not in ag.extra_attrs:
        ag.extra_attrs['aws'] = {}
    if 'key-pair' not in ag.extra_attrs['aws']:
        key_name = 'kraken-%s' % ag.name
        try:
            key_pair = ec2.create_key_pair(KeyName=key_name)
        except botocore.exceptions.ClientError as ex:
            if ex.response['Error']['Code'] == 'InvalidKeyPair.Duplicate':
                ec2.delete_key_pair(KeyName=key_name)
                key_pair = ec2.create_key_pair(KeyName=key_name)
            else:
                log.exception('problem with creating AWS key pair')
                return
        ag.extra_attrs['aws']['key-pair'] = key_pair
        flag_modified(ag, 'extra_attrs')
        db.session.commit()
    else:
        key_name = ag.extra_attrs['aws']['key-pair']['KeyName']

    # prepare security group
    grp_name = 'kraken-%s' % ag.name
    sec_grp = None
    try:
        sec_grp = ec2.describe_security_groups(GroupNames=[grp_name])
    except Exception:
        log.exception('IGNORED EXCEPTION')
    if not sec_grp:
        rsp = requests.get('https://checkip.amazonaws.com')
        my_ip = rsp.text.strip()

        default_vpc = list(ec2_res.vpcs.filter(Filters=[{'Name': 'isDefault', 'Values': ['true']}]))[0]
        sec_grp = default_vpc.create_security_group(GroupName=grp_name,
                                                    Description='kraken sec group that allows only incomming ssh')
        ip_perms = [{
            # SSH ingress open to only the specified IP address
            'IpProtocol': 'tcp', 'FromPort': 22, 'ToPort': 22,
            'IpRanges': [{'CidrIp': '%s/32' % my_ip}]}]
        sec_grp.authorize_ingress(IpPermissions=ip_perms)
        sec_grp_id = sec_grp.id
    else:
        sec_grp_id = sec_grp['SecurityGroups'][0]['GroupId']

    # get AMI ID
    ami_id = system.name

    # define tags
    tags = [{'ResourceType': 'instance',
             'Tags': [{'Key': 'kraken-group', 'Value': '%d' % ag.id}]}]

    # prepare init script
    init_script = """#!/usr/bin/env bash
                     exec > >(tee /var/log/user-data.log|logger -t user-data -s 2>/dev/console) 2>&1
                     wget -O agent {server_url}/bk/install/agent
                     chmod a+x agent
                     ./agent install -s {server_url} -c {clickhouse_addr} --system-id {system_id}
                  """
    init_script = init_script.format(server_url=server_url, clickhouse_addr=clickhouse_addr,
                                     system_id=system.id)
    if aws.get('init_script', False):
        init_script += aws['init_script']

    # CPU Credits spec
    if aws.get('cpu_credits_unlimited', False):
        cpu_credits = 'unlimited'
    else:
        cpu_credits = 'standard'

    # spot instance
    instance_market_opts = {}
    if aws.get('spot_instance', False):
        instance_market_opts = {
            'MarketType': 'spot',
        }

    # collect all params
    params = dict(ImageId=ami_id,
                  MinCount=num,
                  MaxCount=num,
                  KeyName=key_name,
                  SecurityGroupIds=[sec_grp_id],
                  TagSpecifications=tags,
                  InstanceType=aws['instance_type'],
                  Monitoring={'Enabled': aws.get('monitoring', False)},
                  InstanceMarketOptions=instance_market_opts,
                  UserData=init_script)

    if 't2' in aws['instance_type'] or 't3' in aws['instance_type']:
        params['CreditSpecification'] = {'CpuCredits': cpu_credits}

    # prepare disk size if provided
    disk_size = aws.get('disk_size', 0)
    if disk_size:
        params['BlockDeviceMappings'] = [{
            'DeviceName': '/dev/sda1',
            'Ebs': {
                'VolumeSize': disk_size
            }
        }]

    # create AWS EC2 instances
    instances = ec2_res.create_instances(**params)

    log.info('spawning new EC2 VMs for agents %s', instances)

    sys_id = system.id if system.executor == 'local' else 0

    print('instances', instances)
    for i in instances:
        print('iii', i)

        try:
            i.wait_until_running()
        except Exception:
            log.exception('IGNORED EXCEPTION')
            continue
        i.load()
        name = '.'.join(i.public_dns_name.split('.')[:2])
        address = i.private_ip_address
        params = dict(name=name,
                      address=address,
                      ip_address=i.public_ip_address,
                      extra_attrs=dict(system=sys_id, instance_id=i.id))
        a = _create_agent(params, ag)
        log.info('spawned new agent %s on EC2 instance %s', a, i)


def destroy_vm(ag, agent):  # pylint: disable=unused-argument
    credential = login_to_aws()
    if not credential:
        return

    _, depl = ag.get_deployment()

    region = depl['region']
    ec2 = boto3.resource('ec2', region_name=region, **credential)

    instance_id = agent.extra_attrs['instance_id']
    log.info('terminate ec2 vm %s', instance_id)
    try:
        i = ec2.Instance(instance_id)
        i.terminate()
    except Exception:
        log.exception('IGNORED EXCEPTION')


def vm_exists(ag, agent):
    credential = login_to_aws()
    if not credential:
        raise Exception('wrong aws credential')

    _, depl = ag.get_deployment()

    region = depl['region']
    ec2 = boto3.resource('ec2', region_name=region, **credential)

    instance_id = agent.extra_attrs['instance_id']
    try:
        # try to get instance, if missing then raised exception will cause return False
        i = ec2.Instance(instance_id)
        i.state  # pylint: disable=pointless-statement
        if i.state['Name'] == 'terminated':
            # if instance exists theb raising an exception and it will cause return False
            raise Exception('terminated')
    except Exception:
        return False

    return True


def cleanup_dangling_vms(ag):
    credential = login_to_aws()
    if not credential:
        return 0, 0, 0, 0, 0

    _, depl = ag.get_deployment()

    region = depl['region']
    ec2 = boto3.resource('ec2', region_name=region, **credential)

    now = utils.utcnow()

    try:
        vms = ec2.instances.filter(Filters=[{'Name': 'tag:kraken-group', 'Values': ['%d' % ag.id]}])
        vms = list(vms)
    except Exception:
        log.exception('IGNORED EXCEPTION')
        return 0, 0, 0, 0, 0

    instances = 0
    terminated_instances = 0
    assigned_instances = 0
    orphaned_instances = 0
    orphaned_terminated_instances = 0

    for vm in vms:
        instances += 1
        # if terminated then skip it
        if vm.state['Name'] == 'terminated':
            terminated_instances += 1
            continue

        # if assigned to some agent then skip it
        assigned = False
        for aa in ag.agents:
            agent = aa.agent
            if agent.extra_attrs and 'instance_id' in agent.extra_attrs and agent.extra_attrs['instance_id'] == vm.id:
                assigned = True
                break
        if assigned:
            assigned_instances += 1
            continue

        # instances have to be old enough to avoid race condition with
        # case when instances are being created but not yet assigned to agents
        lt = vm.launch_time.replace(tzinfo=pytz.utc)
        if now - lt < datetime.timedelta(minutes=10):
            continue

        # the instance is not terminated, not assigned, old enough
        # so delete it as it seems to be a lost instance
        log.info('terminating lost aws ec2 instance %s', vm.id)
        orphaned_instances += 1
        try:
            vm.terminate()
        except Exception:
            log.exception('IGNORED EXCEPTION')

        orphaned_terminated_instances += 1

    return instances, terminated_instances, assigned_instances, orphaned_instances, orphaned_terminated_instances



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from flask import send_file, abort, make_response

from .models import get_setting

log = logging.getLogger(__name__)


KKAGENT_DIR = os.environ.get('KKAGENT_DIR', '')


INSTALL_SCRIPT = '''#!/bin/bash
set -x

KRAKEN_URL="{url}"

# check sudo
sudo -n true
if [ $? -ne 0 ]; then
    echo "error: sudo with no password is required"
    exit 1
fi

# check python3
python3 --version
if [ $? -ne 0 ]; then
    echo "error: missing python3, it is required to run Kraken agent"
    exit 1
fi

# check curl and wget
DL_TOOL=
which curl
if [ $? -ne 0 ]; then
    which wget
    if [ $? -ne 0 ]; then
        echo "error: missing curl and wget, install any of them"
        exit 1
    else
        DL_TOOL=wget
    fi
else
    DL_TOOL=curl
fi

set -e

if [ "$DL_TOOL" == "wget" ]; then
    wget -O /tmp/kkagent ${KRAKEN_URL}/install/agent
else
    curl -o /tmp/kkagent ${KRAKEN_URL}/install/agent
fi
chmod a+x /tmp/kkagent

export KRAKEN_CLICKHOUSE_ADDR="{clickhouse-addr}"

/tmp/kkagent install -s ${KRAKEN_URL}
rm -f /tmp/kkagent
echo 'Kraken Agent installed'
'''


def serve_agent_blob(blob):
    if blob not in ['agent', 'tool', 'kraken-agent-install.sh']:
        abort(404)

    if blob == 'kraken-agent-install.sh':
        # get and check server url
        url = get_setting('general', 'server_url')
        if not url:
            abort(500, 'Cannot get server URL and put it in Kraken agent install script. ' +
                  'The URL should be set on Kraken Settings page first.')

        # get and check clickhouse addr
        clickhouse_addr = get_setting('general', 'clickhouse_addr')
        if not clickhouse_addr:
            abort(500, 'Cannot get ClickHouse address and put it in Kraken agent install script. ' +
                  'The ClickHouse address should be set on Kraken Settings page first.')

        # patch install script with url and addresses
        script = INSTALL_SCRIPT.replace('{url}', url)
        script = script.replace('{clickhouse-addr}', clickhouse_addr)

        resp = make_response(script)
        # add a filename
        resp.headers.set("Content-Type", "application/x-sh")
        resp.headers.set("Content-Disposition", "attachment", filename="kraken-agent-install.sh")
        return resp

    p = os.path.join(KKAGENT_DIR, 'kk' + blob)
    return send_file(p)



#!/usr/bin/env python3

# Copyright 2021 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import logging

import rq
import redis
from flask import Flask

from . import srvcheck
from . import consts
from . import logs
from . import utils
from .. import version
from .models import db, get_setting

log = logging.getLogger('rq')


def enq(func, *args, **kwargs):
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_RQ_DB)
    q = rq.Queue('kq', connection=rds)
    # job timeout is 20mins (1200s), interval between retries is 60s
    j = q.enqueue(func, args=args, kwargs=kwargs, retry=rq.Retry(max=2, interval=5), job_timeout=1200)
    log.info('put to bg processing: %s', j)
    return j


def enq_neck(func, *args, ignore_args=None):
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_RQ_DB)
    data = dict(func=func.__name__,
                args=args,
                ignore_args=ignore_args)
    data = json.dumps(data)
    rds.publish('qneck', data)
    log.info('enqueued func %s with args: %s and %s', func.__name__, args, ignore_args)


def get_jobs():
    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_RQ_DB)
    q = rq.Queue('kq', connection=rds)

    jobs_ids = q.scheduled_job_registry.get_job_ids()
    scheduled_jobs = rq.job.Job.fetch_many(jobs_ids, connection=rds)

    jobs_ids = q.started_job_registry.get_job_ids()
    started_jobs = rq.job.Job.fetch_many(jobs_ids, connection=rds)

    jobs_ids = q.finished_job_registry.get_job_ids()
    finished_jobs = rq.job.Job.fetch_many(jobs_ids, connection=rds)

    jobs_ids = q.failed_job_registry.get_job_ids()
    failed_jobs = rq.job.Job.fetch_many(jobs_ids, connection=rds)

    jobs_ids = q.deferred_job_registry.get_job_ids()
    deferred_jobs = rq.job.Job.fetch_many(jobs_ids, connection=rds)

    return scheduled_jobs, started_jobs, finished_jobs, failed_jobs, deferred_jobs


# TODO: it does not work because rq always traces the exception
# def _exception_handler(job, exc_type, exc_value, traceback):  # pylint: disable=unused-argument
#     ignore_excs = ['ix_agents_address']
#     for ex in ignore_excs:
#         if ex in str(exc_value) or ex in str(traceback):
#             return
#     log.exception('IGNORED')


def main():
    # check deps
    planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)
    srvcheck.wait_for_service('planner', planner_url, 7997)

    redis_addr = os.environ.get('KRAKEN_REDIS_ADDR', consts.DEFAULT_REDIS_ADDR)
    srvcheck.wait_for_service('redis', redis_addr, 6379)
    redis_host, redis_port = utils.split_host_port(redis_addr, 6379)
    rds = redis.Redis(host=redis_host, port=redis_port, db=consts.REDIS_RQ_DB)

    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    srvcheck.check_postgresql(db_url)

    logs.setup_logging('rq')
    log.info('Kraken RQ started, version %s', version.version)

    # Create Flask app instance
    app = Flask('Kraken RQ')
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url + '?application_name=rq'
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    db.init_app(app)

    # setup sentry
    with app.app_context():
        sentry_url = get_setting('monitoring', 'sentry_dsn')
        logs.setup_sentry(sentry_url)

    worker = rq.Worker('kq', connection=rds)  # , exception_handlers=[_exception_handler])  TODO: see above
    worker.work(with_scheduler=True)


if __name__ == "__main__":
    main()



"""added unique constraint for system name and executor

Revision ID: b2c9024746f5
Revises: 1f5039135dd8
Create Date: 2021-06-20 12:17:02.966101

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = 'b2c9024746f5'
down_revision = '1f5039135dd8'
branch_labels = None
depends_on = None


def upgrade():
    with op.get_context().autocommit_block():
        try:
            op.create_unique_constraint('uq_system_name_executor', 'systems', ['name', 'executor'])
        except Exception:
            pass


def downgrade():
    op.drop_constraint('uq_system_name_executor', 'systems', type_='unique')



"""add extra to Issue

Revision ID: b5cc62f79980
Revises: 1cf31ab18f84
Create Date: 2019-12-26 11:17:33.681018

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'b5cc62f79980'
down_revision = '1cf31ab18f84'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('issues', sa.Column('extra', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('issues', 'extra')



"""changed tool version to string

Revision ID: befa07ac5d1c
Revises: af348b43d9d5
Create Date: 2022-05-26 07:18:25.697901

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'befa07ac5d1c'
down_revision = 'af348b43d9d5'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('tools', 'version',
               existing_type=sa.INTEGER(),
               type_=sa.UnicodeText(),
               existing_nullable=True)
    op.create_unique_constraint('uq_tool_name_version', 'tools', ['name', 'version'])

    conn = op.get_bind()
    conn.execute("UPDATE tools SET version = '1'")


def downgrade():
    op.drop_constraint('uq_tool_name_version', 'tools', type_='unique')
    op.alter_column('tools', 'version',
               existing_type=sa.UnicodeText(),
               type_=sa.INTEGER(),
               existing_nullable=True)



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy.sql.expression import desc

from .models import Run, Flow, db
from . import utils
from . import consts


def get_prev_run(stage_id, flow_kind):
    q = Run.query
    q = q.filter_by(deleted=None)
    q = q.filter_by(stage_id=stage_id)
    q = q.join('flow')
    q = q.filter(Flow.kind == flow_kind)
    q = q.order_by(desc(Flow.created))
    prev_run = q.first()
    return prev_run


def find_cloud_assignment_group(agent):
    for aa in agent.agents_groups:
        ag = aa.agents_group
        if ag.deployment:
            return ag
    return None


def delete_agent(agent):
    agent.deleted = utils.utcnow()
    agent.disabled = True
    agent.authorized = False
    for aa in agent.agents_groups:
        db.session.delete(aa)
    db.session.commit()


def get_secret_values(project):
    secrets = []
    for s in project.secrets:
        if s.deleted:
            continue
        if s.kind == consts.SECRET_KIND_SSH_KEY:
            secrets.append(s.data['key'])
        elif s.kind == consts.SECRET_KIND_SIMPLE:
            secrets.append(s.data['secret'])
    return secrets



"""added unique constraints on branch name and project

Revision ID: e404f52562c2
Revises: 42ea01a6cf31
Create Date: 2021-06-27 16:00:20.034147

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e404f52562c2'
down_revision = '42ea01a6cf31'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('ix_branches_name_project_id_not_deleted', 'branches', ['name', 'project_id'], unique=True, postgresql_where=sa.text('deleted IS NULL'))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('ix_branches_name_project_id_not_deleted', table_name='branches')
    # ### end Alembic commands ###



"""added processed_at in run and index to tcr

Revision ID: a35aec7afa0a
Revises: 54efe873f88f
Create Date: 2022-01-01 09:29:44.529228

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a35aec7afa0a'
down_revision = '54efe873f88f'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('runs', sa.Column('processed_at', sa.DateTime(timezone=True), nullable=True))
    op.create_index('ix_test_case_results_test_case_id', 'test_case_results', ['test_case_id'], unique=False)


def downgrade():
    op.drop_index('ix_test_case_results_test_case_id', table_name='test_case_results')
    op.drop_column('runs', 'processed_at')



#!/usr/bin/env python3

# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import logging

from flask import Flask
from sqlalchemy.sql.expression import asc

from . import logs
from .models import db, Agent, Run, Job, get_setting
from . import consts
from . import srvcheck
from .. import version
from . import utils

log = logging.getLogger('scheduler')


def _get_idle_agents():
    all_idle_agents = Agent.query.filter_by(job=None, authorized=True, disabled=False, deleted=None).all()
    agents_count = len(all_idle_agents)
    if agents_count == 0:
        return 0, {}, {}

    idle_agents_by_group = {}
    idle_agents_by_sys_group = {}
    for ag in all_idle_agents:
        log.info('idle agent: %s', ag)
        for asm in ag.agents_groups:
            grp_id = asm.agents_group_id
            log.info('  grp: %s', grp_id)
            if grp_id not in idle_agents_by_group:
                idle_agents_by_group[grp_id] = []
            idle_agents_by_group[grp_id].append(ag)

            sys_grp_key = (asm.agents_group_id,
                           ag.host_info['system'] if ag.host_info and 'system' in ag.host_info else 'fake')
            if sys_grp_key not in idle_agents_by_sys_group:
                idle_agents_by_sys_group[sys_grp_key] = []
            idle_agents_by_sys_group[sys_grp_key].append(ag)

    return agents_count, idle_agents_by_group, idle_agents_by_sys_group


def _get_waiting_jobs():
    q = Job.query.filter_by(state=consts.JOB_STATE_QUEUED, agent_used=None)
    q = q.join('run')
    waiting_jobs = q.order_by(asc(Run.created), asc(Job.created)).all()  # FIFO

    return waiting_jobs


def _assign_jobs(agents_count, idle_agents_by_group, idle_agents_by_sys_group,
                 waiting_jobs):
    counter = 0

    for j in waiting_jobs:
        log.set_ctx(branch=j.run.flow.branch_id, flow_kind=j.run.flow.kind, flow=j.run.flow_id, run=j.run.id)
        log.info('job %s, executor %s', j, j.system.executor)

        # system does not have to be taken into account when user selects ANY system
        # or when executor is not local eg. it is a docker container
        if j.system.name == 'any' or j.system.executor != 'local':
            idle_agents = idle_agents_by_group.get(j.agents_group_id, [])
        else:
            key = (j.agents_group_id, j.system.name)
            idle_agents = idle_agents_by_sys_group.get(key, [])

        # find idle agent from given agents groups
        best_agent = None
        while best_agent is None:
            if len(idle_agents) == 0:
                log.info('no avail agents for job %s', j)
                break
            best_agent = idle_agents.pop()
            if best_agent.job:
                log.info('agent busy %s', best_agent)
                best_agent = None

        if best_agent is None:
            log.info('no agents for job %s', j)
            continue

        # assign job to found agent
        best_agent.job = j
        j.agent_used = best_agent
        j.assigned = utils.utcnow()
        db.session.commit()
        log.info("assigned job %s to agent %s", j, best_agent)

        counter += 1
        # if all idle agents used then stop assigning
        if counter == agents_count:
            break

    log.reset_ctx()

    return counter


def assign_jobs_to_agents():

    agents_count, idle_agents_by_group, idle_agents_by_sys_group = _get_idle_agents()
    if agents_count == 0:
        log.info('no idle agents')
        return 0

    waiting_jobs = _get_waiting_jobs()

    if len(waiting_jobs) > 0:
        log.info('idle agents: %s, waiting jobs %s', agents_count, waiting_jobs)

    counter = _assign_jobs(agents_count, idle_agents_by_group, idle_agents_by_sys_group, waiting_jobs)

    return counter


def create_app():
    # addresses
    db_url = os.environ.get('KRAKEN_DB_URL', consts.DEFAULT_DB_URL)
    planner_url = os.environ.get('KRAKEN_PLANNER_URL', consts.DEFAULT_PLANNER_URL)

    srvcheck.check_postgresql(db_url)
    srvcheck.wait_for_service('planner', planner_url, 7997)

    logs.setup_logging('scheduler')
    log.info('Kraken Scheduler started, version %s', version.version)

    # Create  Flask app instance
    app = Flask('Kraken Scheduler')

    # Configure the SqlAlchemy part of the app instance
    app.config["SQLALCHEMY_ECHO"] = False
    app.config["SQLALCHEMY_DATABASE_URI"] = db_url + '?application_name=scheduler'
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False

    # initialize SqlAlchemy
    db.init_app(app)

    # setup sentry
    with app.app_context():
        sentry_url = get_setting('monitoring', 'sentry_dsn')
        logs.setup_sentry(sentry_url)

    return app


def main():
    app = create_app()

    with app.app_context():

        while True:
            t0 = time.time()
            jobs_cnt = assign_jobs_to_agents()
            t1 = time.time()

            dt = t1 - t0
            sleep_time = 5 - dt
            sleep_time = max(sleep_time, 0)
            log.info("scheduled %d jobs in %.1fs, go sleep for %.1fs", jobs_cnt, dt, sleep_time)
            if sleep_time > 0:
                time.sleep(sleep_time)


if __name__ == "__main__":
    main()



"""added group to Setting

Revision ID: 8b4adeca953d
Revises: 8200318b9e18
Create Date: 2020-02-08 12:44:36.855759

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8b4adeca953d'
down_revision = '8200318b9e18'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('settings', sa.Column('group', sa.VARCHAR(length=50), autoincrement=False, nullable=True))


def downgrade():
    op.drop_column('settings', 'group')



# Copyright 2020-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from passlib.hash import pbkdf2_sha256

from .models import db, Branch, Stage, AgentsGroup, Setting, Tool, Project
from .models import User, BranchSequence, System
from .schema import execute_schema_code
from . import consts


# pylint: disable=line-too-long


INITIAL_SETTINGS = {
    'general': {
        'server_url': '',
        'minio_addr': '',
        'clickhouse_addr': '',
        'clickhouse_log_ttl': 6,
    },
    'notification': {
        'smtp_server': '',
        'smtp_tls': False,
        'smtp_from': '',
        'smtp_user': '',
        'smtp_password': None,  # password
        'slack_token': None  # password
    },
    'monitoring': {
        'sentry_dsn': None  # password
    },
    'cloud': {
        # AWS
        'aws_access_key': '',
        'aws_secret_access_key': None,  # password
        # Azure
        'azure_subscription_id': '',
        'azure_tenant_id': '',
        'azure_client_id': '',
        'azure_client_secret': None,  # password
        # Kubernetes
        'k8s_api_server_url': '',
        'k8s_namespace': 'kraken',
        'k8s_token': None,  # password
    },
    'idp': {
        # LDAP
        'ldap_enabled': False,
        'ldap_server': '',
        'bind_dn': '',
        'bind_password': None,  # password
        'base_dn': '',
        'search_filter': '',

        # Google OIDC
        'google_enabled': False,
        'google_client_id': '',
        'google_client_secret': None,  # password

        # Microsoft Azure
        'microsoft_enabled': False,
        'microsoft_client_id': '',
        'microsoft_client_secret': None,  # password

        # GitHub
        'github_enabled': False,
        'github_client_id': '',
        'github_client_secret': None,  # password

        # Auth0
        'auth0_enabled': False,
        'auth0_client_id': '',
        'auth0_client_secret': None,  # password
        'auth0_openid_config_url': '',
    }
}


def _prepare_initial_preferences():
    for group_name, group_fields in INITIAL_SETTINGS.items():
        for name, val in group_fields.items():
            s = Setting.query.filter_by(group=group_name, name=name).one_or_none()
            if s is not None:
                # correct type if necessary
                if isinstance(val, bool):
                    s.val_type = 'boolean'
                elif isinstance(val, int):
                    s.val_type = 'integer'
                elif val is None:
                    s.val_type = 'password'
                else:
                    s.val_type = 'text'
                continue

            # set value and type
            if isinstance(val, bool):
                val_type = 'boolean'
                val = str(val)
            elif isinstance(val, int):
                val_type = 'integer'
                val = str(val)
            elif val is None:
                val_type = 'password'
                val = ''
            else:
                val_type = 'text'
            Setting(group=group_name, name=name, value=val, val_type=val_type)
    db.session.commit()


def _prepare_builtin_tools():
    tool_defs = [{
        "name": "local_tool",
        "description": "A tool that allows for running arbitrary python script as a tool that is indicated by `tool_location` and `tool_entry` fields. It is possible to add arbitrary fields to step definition that will be consumed by this tool.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "additionalProperties": True,
            "required": ["tool_location", "tool_entry"],
            "properties": {
                "tool_location": {
                    "description": "A folder where a Python script is located.",
                    "type": "string"
                },
                "tool_entry": {
                    "description": "A Python script module name i.e. file name without `.py` suffix.",
                    "type": "string"
                },
            }
        }
    }, {
        "name": "git",
        "description": "A tool for cloning Git repository.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "additionalProperties": False,
            "required": ["checkout"],
            "properties": {
                "checkout": {
                    "description": "An URL to the repository.",
                    "type": "string"
                },
                "branch": {
                    "description": "A branch to checkout.",
                    "default": "master",
                    "type": "string"
                },
                "destination": {
                    "description": "A destination folder for the repository. Default is empty ie. the name of the repository.",
                    "type": "string"
                },
                "ssh-key": {
                    "description": "A name of a secret that holds SSH username and key.",
                    "type": "string"
                },
                "access-token": {
                    "description": "A name of secret that contains an access token for GitLab or GitHub.",
                    "type": "string"
                },
                "timeout": {
                    "description": "A timeout in seconds that limits time of step execution. It is guareded by an agent. If it is exceeded then the step is arbitrarly terminated.",
                    "type": "integer",
                    "minimum": 30
                },
                "git_cfg": {
                    "description": "Git config keys and values passed to -c of the clone command.",
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                }
            }
        }
    }, {
        "name": "shell",
        "description": "A tool that executes provided command in a shell.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "properties": {
                "cmd": {
                    "description": "A command to execute.",
                    "type": "string"
                },
                "script": {
                    "description": "A script code to execute.",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                },
                "user": {
                    "description": "A user that is used to execute a command.",
                    "default": "kraken",
                    "type": "string"
                },
                "env": {
                    "description": "A dictionary with environment variables and their values.",
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                },
                "timeout": {
                    "description": "A timeout in seconds that limits time of step execution. It is guareded by an agent. If it is exceeded then the step is arbitrarly terminated.",
                    "type": "integer",
                    "minimum": 30,
                    "default": 60
                },
                "background": {
                    "description": "Indicates if step should be started and pushed to background. The step process is closed at the end of a job.",
                    "default": False,
                    "type": "boolean"
                },
                "shell_exe": {
                    "description": "An alternative path or command to shell executable (e.g.: zsh or /usr/bin/fish).",
                    "type": "string"
                },
                "ssh-key": {
                    "description": "A name of a secret that holds SSH username and key.",
                    "type": "string"
                },
            }
        }
    }, {
        "name": "pytest",
        "description": "A tool that allows for running Python tests.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "properties": {
                "pytest_exe": {
                    "description": "An alternative path or command to pytest.",
                    "default": "pytest-3",
                    "type": "string"
                },
                "params": {
                    "description": "Parameters passed directly to pytest executable.",
                    "type": "string"
                },
                "pythonpath": {
                    "description": "Extra paths that are used by Python to look for modules/packages that it wants to load.",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                }
            }
        }
    }, {
        "name": "rndtest",
        "description": "A tool that allows for generating random test case results.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "properties": {
                "count": {
                    "description": "A number of expected test cases.",
                    "oneOf": [{
                        "type": "integer",
                        "minimum": 1
                    }, {
                        "type": "string"
                    }]
                },
                "override_result": {
                    "description": "A result.",
                    "oneOf": [{
                        "type": "integer",
                        "minimum": 1
                    }, {
                        "type": "string"
                    }]
                },
            }
        }
    }, {
        "name": "artifacts",
        "description": "A tool for storing and retrieving artifacts in Kraken global storage.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "required": ["source"],
            "properties": {
                "action": {
                    "description": "An action that artifacts tool should execute. Default is `upload`.",
                    "type": "string",
                    "enum": ["download", "upload"]
                },
                "source": {
                    "description": "A path or list of paths that should be archived or retreived. A path can indicate a folder or a file. A path, in case of upload action, can contain globbing signs `*` or `**`. A path can be relative or absolute.",
                    "oneOf": [{
                        "description": "A single path.",
                        "type": "string"
                    }, {
                        "description": "A list of paths.",
                        "type": "array",
                        "items": {
                            "type": "string"
                        }
                    }]
                },
                "destination": {
                    "description": "A path were the artifact(s) should be stored. In case of download action, if the destination folder does not exist then it is created.",
                    "default": ".",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                },
                "public": {
                    "description": "Determines if artifacts should be public and available to users in web UI (`True`) or if they should be only accessible internally to other stages but only in the same flow (`False`). If report_entry is set then public is True.",
                    "default": False,
                    "type": "boolean"
                },
                "report_entry": {
                    "description": "A path to HTML file that is an entry to the uploaded report. If present then it sets public to True.",
                    "type": "string"
                },
            }
        }
    }, {
        "name": "pylint",
        "description": "A tool that allows for static analysis of Python source code.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "required": ["modules_or_packages"],
            "properties": {
                "pylint_exe": {
                    "description": "An alternative path or command to pylint.",
                    "default": "pylint",
                    "type": "string"
                },
                "rcfile": {
                    "description": "A path to pylint rcfile.",
                    "type": "string"
                },
                "modules_or_packages": {
                    "description": "A path or paths to Python modules or packages that should be checked.",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                },
                "timeout": {
                    "description": "A timeout in seconds that limits time of step execution. It is guareded by an agent. If it is exceeded then the step is arbitrarly terminated.",
                    "type": "integer",
                    "minimum": 30
                }
            }
        }
    }, {
        "name": "cloc",
        "description": "A tool that allows for running counting lines of code.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "properties": {
                "not-match-f": {
                    "description": "Filter out files that match to provided regex.",
                    "type": "string"
                },
                "exclude-dir": {
                    "description": "Excluded provided list of directories.",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                },
            }
        }
    }, {
        "name": "nglint",
        "description": "A tool that allows for running Angular `ng lint`, that is performing static analysis of TypeScript in Angular projects.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "properties": {
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                },
            }
        }
    }, {
        "name": "cache",
        "description": "A tool for storing and restoring files from cache.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "required": ["action"],
            "properties": {
                "action": {
                    "description": "An action that the tool should perform.",
                    "type": "string",
                    "enum": ["save", "restore"]
                },
                "key": {
                    "description": "A key under which files are stored in or restored from cache.",
                    "type": "string"
                },
                "keys": {
                    "description": "A list of key under which files are restored from cache.",
                    "type": "array",
                    "items": {
                        "type": "string"
                    }
                },
                "paths": {
                    "description": "Source paths used in `store` action.",
                    "type": "array",
                    "items": {
                        "type": "string"
                    }
                },
                "expiry": {
                    "description": "Not implemented yet.",
                    "type": "string"
                },
            }
        }
    }, {
        "name": "gotest",
        "description": "A tool that allows for running Go language tests.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "properties": {
                "go_exe": {
                    "description": "An alternative path or command to `go`.",
                    "type": "string"
                },
                "params": {
                    "description": "Parameters passed directly to `go test`.",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                },
                "timeout": {
                    "description": "A timeout in seconds that limits time of step execution. It is guareded by an agent. If it is exceeded then the step is arbitrarly terminated.",
                    "type": "integer",
                    "minimum": 30
                },
            }
        }
    }, {
        "name": "junit_collect",
        "description": "A tool that allows for collecting test results stored in JUnit files.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "required": ["file_glob"],
            "properties": {
                "file_glob": {
                    "description": "A glob pattern for searching test result files.",
                    "type": "string"
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                }
            }
        }
    }, {
        "name": "values_collect",
        "description": "A tool that allows for collecting values (metrics, params, etc) from files.",
        "location": "",
        "entry": "",
        "version": "1",
        "parameters": {
            "required": ["files"],
            "properties": {
                "files": {
                    "description": "A list of files.",
                    "type": "array",
                    "items": {
                        "type": "object",
                        "additionalProperties": False,
                        "properties": {
                            "name": {
                                "description": ".",
                                "type": "string"
                            },
                            "namespace": {
                                "description": ".",
                                "type": "string"
                            }
                        }
                    }
                },
                "cwd": {
                    "description": "A current working directory where the step is executed.",
                    "default": ".",
                    "type": "string"
                }
            }
        }
    }]
    for td in tool_defs:
        tool = Tool.query.filter_by(name=td['name'], version=td['version']).one_or_none()
        if tool is None:
            tool = Tool(name=td['name'],
                        description=td['description'],
                        location=td['location'],
                        entry=td['entry'],
                        version=td['version'],
                        fields=td['parameters'])
            print("   created Tool record", td['name'])
        else:
            tool.name = td['name']
            tool.description = td['description']
            tool.location = td['location']
            tool.entry = td['entry']
            tool.version = td['version']
            tool.fields = td['parameters']
            print("   updated Tool record", td['name'])
        db.session.commit()


def prepare_initial_data():
    print("Preparing initial DB data")

    _prepare_builtin_tools()

    agents_group = AgentsGroup.query.filter_by(name="all").one_or_none()
    if agents_group is None:
        AgentsGroup(name='all')
        db.session.commit()
        print("   created AgentsGroup record 'all'")

        # AgentAssignment(agent=agent, agents_group=agents_group)
        # db.session.commit()
        # print("   created AgentAssignment for record 'all'")

    # Project DEMO
    project = Project.query.filter_by(name="Demo").one_or_none()
    if project is None:
        project = Project(name='Demo', description="This is a demo project.")
        db.session.commit()
        print("   created Project record 'Demo'")

    branch = Branch.query.filter_by(name="Master", project=project).one_or_none()
    if branch is None:
        branch = Branch(name='Master', branch_name='master', project=project)
        db.session.commit()
        print("   created Branch record 'master'")

    for seq_type in [consts.BRANCH_SEQ_FLOW, consts.BRANCH_SEQ_CI_FLOW, consts.BRANCH_SEQ_DEV_FLOW]:
        bs = BranchSequence.query.filter_by(branch=branch, kind=seq_type).one_or_none()
        if bs is None:
            BranchSequence(branch=branch, kind=seq_type, value=0)
    db.session.commit()

    stage = Stage.query.filter_by(name="Tests", branch=branch).one_or_none()
    schema_code = '''def stage(ctx):
    return {
        "parent": "root",
        "triggers": {
            "parent": True
        },
        "parameters": [],
        "configs": [],
        "jobs": [{
            "name": "test",
            "steps": [{
                "tool": "git",
                "checkout": "https://github.com/Kraken-CI/sample-project-python.git",
                "branch": "master"
            }, {
                "tool": "pytest",
                "cwd": "sample-project-python",
                "params": "-vv",
                "pythonpath": "src"
            }],
            "environments": [{
                "system": "any",
                "agents_group": "all",
                "config": "default"
            }]
        }]
    }'''
    if stage is None:
        stage = Stage(name='Tests', description="This is a stage of tests.", branch=branch,
                      schema_code=schema_code, schema=execute_schema_code(branch, schema_code))
        db.session.commit()
        print("   created Stage record 'Tests'")
    else:
        stage.schema_code = schema_code
        stage.schema = execute_schema_code(branch, schema_code)
        db.session.commit()


    for seq_type in [consts.BRANCH_SEQ_RUN, consts.BRANCH_SEQ_CI_RUN, consts.BRANCH_SEQ_DEV_RUN]:
        bs = BranchSequence.query.filter_by(branch=branch, stage=stage, kind=seq_type).one_or_none()
        if bs is None:
            BranchSequence(branch=branch, stage=stage, kind=seq_type, value=0)
    db.session.commit()

    # create default users: admin and demo

    admin_user = User.query.filter_by(name="admin").one_or_none()
    if admin_user is None:
        password = pbkdf2_sha256.hash('admin')
        User(name='admin', password=password)
        db.session.commit()
        print("   created User record 'admin'")

    demo_user = User.query.filter_by(name="demo").one_or_none()
    if demo_user is None:
        password = pbkdf2_sha256.hash('demo')
        User(name='demo', password=password)
        db.session.commit()
        print("   created User record 'demo'")

    # common systems
    system = System.query.filter_by(name='any', executor='local').one_or_none()
    if system is None:
        System(name='any', executor='local')
        db.session.commit()

    # preferences
    _prepare_initial_preferences()



"""added Artifact and File

Revision ID: bb0d4908ba49
Revises: 2b5751b83fec
Create Date: 2020-02-16 07:42:24.358297

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'bb0d4908ba49'
down_revision = '2b5751b83fec'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('files',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('path', sa.Unicode(length=512), nullable=True),
                    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('artifacts',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('file_id', sa.Integer(), nullable=False),
                    sa.Column('flow_id', sa.Integer(), nullable=False),
                    sa.Column('run_id', sa.Integer(), nullable=False),
                    sa.Column('size', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(['file_id'], ['files.id'], ),
                    sa.ForeignKeyConstraint(['flow_id'], ['flows.id'], ),
                    sa.ForeignKeyConstraint(['run_id'], ['runs.id'], ),
                    sa.PrimaryKeyConstraint('id')
    )
    op.add_column('runs', sa.Column('artifacts', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('runs', 'artifacts')
    op.drop_table('artifacts')
    op.drop_table('files')



"""added results change stats to Run

Revision ID: 98af95844f27
Revises: 818e190ca142
Create Date: 2020-01-04 18:32:23.969038

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '98af95844f27'
down_revision = '818e190ca142'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('runs', sa.Column('fix_cnt', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('new_cnt', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('no_change_cnt', sa.Integer(), nullable=True))
    op.add_column('runs', sa.Column('regr_cnt', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('runs', 'regr_cnt')
    op.drop_column('runs', 'no_change_cnt')
    op.drop_column('runs', 'new_cnt')
    op.drop_column('runs', 'fix_cnt')



# Copyright 2020-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from flask_sqlalchemy import SQLAlchemy
from sqlalchemy import Column, Boolean, DateTime, ForeignKey, Integer, Unicode, UnicodeText, String
from sqlalchemy import event, UniqueConstraint, Index
from sqlalchemy.orm import relationship, mapper
from sqlalchemy.dialects.postgresql import JSONB, DOUBLE_PRECISION, BYTEA

from . import consts
from . import utils

log = logging.getLogger(__name__)

db = SQLAlchemy(engine_options=dict(connect_args={"options": "-c timezone=utc"}))


@event.listens_for(mapper, 'init')
def auto_add(target, args, kwargs):  # pylint: disable=unused-argument
    db.session.add(target)


def duration_to_txt(duration):
    duration_txt = ""
    if duration.days > 0:
        duration_txt = "%dd " % duration.days
    if duration.seconds > 3600:
        duration_txt += "%dh " % (duration.seconds // 3600)
    if duration.seconds > 60:
        seconds = duration.seconds % 3600
        duration_txt += "%dm " % (seconds // 60)
    duration_txt += "%ds" % (duration.seconds % 60)
    duration_txt = duration_txt.strip()
    return duration_txt


class AlembicVersion(db.Model):
    __tablename__ = "alembic_version"
    version_num = Column(Unicode(32), nullable=False, primary_key=True)


class DatesMixin():
    created = Column(DateTime(timezone=True), nullable=False, default=utils.utcnow)
    updated = Column(DateTime(timezone=True), nullable=False, default=utils.utcnow, onupdate=utils.utcnow)
    deleted = Column(DateTime(timezone=True))


class Project(db.Model, DatesMixin):
    __tablename__ = "projects"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    description = Column(UnicodeText)
    branches = relationship("Branch", back_populates="project", order_by="Branch.created")
    secrets = relationship("Secret", back_populates="project", order_by="Secret.name")
    agents_groups = relationship("AgentsGroup", back_populates="project")
    webhooks = Column(JSONB)

    def get_json(self, with_results=False, with_last_results=False):
        branches = [b.get_json(with_results=with_results, with_last_results=with_last_results) for b in self.branches if b.deleted is None]
        branches.sort(key=lambda b: b['name'])
        return dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    name=self.name,
                    description=self.description,
                    branches=branches,
                    secrets=[s.get_json() for s in self.secrets if s.deleted is None],
                    webhooks=self.webhooks if self.webhooks else {})


class Branch(db.Model, DatesMixin):
    __tablename__ = "branches"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)  # display name
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)
    project = relationship('Project', back_populates="branches")
    branch_name = Column(UnicodeText)                             # branch name in the repository, PR is matched agains this
    ci_flows = relationship("Flow", order_by="desc(Flow.created)",
                            primaryjoin="and_(Branch.id==Flow.branch_id, Flow.kind==0)", viewonly=True)
    dev_flows = relationship("Flow", order_by="desc(Flow.created)",
                             primaryjoin="and_(Branch.id==Flow.branch_id, Flow.kind==1)", viewonly=True)
    ci_last_completed_flow_id = Column(Integer, ForeignKey('flows.id'))
    ci_last_completed_flow = relationship('Flow', foreign_keys=[ci_last_completed_flow_id])
    ci_last_incomplete_flow_id = Column(Integer, ForeignKey('flows.id'))
    ci_last_incomplete_flow = relationship('Flow', foreign_keys=[ci_last_incomplete_flow_id])
    stages = relationship("Stage", back_populates="branch", order_by="Stage.name")
    sequences = relationship("BranchSequence", back_populates="branch")
    comments = relationship("TestCaseComment", back_populates="branch")
    retention_policy = Column(JSONB)

    #base_branch = relationship('BaseBranch', uselist=False, primaryjoin="or_(Branch.id==BaseBranch.ci_branch_id, Branch.id==BaseBranch.dev_branch_id)")

    def get_json(self, with_results=False, with_cfg=False, with_last_results=False):
        if self.retention_policy:
            retention_policy = self.retention_policy
        else:
            retention_policy = dict(ci_logs=6,
                                    dev_logs=3,
                                    ci_artifacts=6,
                                    dev_artifacts=3)

        data = dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    name=self.name,
                    project_id=self.project_id,
                    project_name=self.project.name,
                    branch_name=self.branch_name,
                    retention_policy=retention_policy)

        if with_results:
            data['ci_flows'] = [f.get_json() for f in self.ci_flows[:10]]
            data['dev_flows'] = [f.get_json() for f in self.dev_flows[:10]]

        if with_last_results:
            data['last_completed_flow'] = None
            data['last_incomplete_flow'] = None

            f = self.ci_last_completed_flow
            if f:
                f_json = dict(id=f.id,
                              label=f.label,
                              finished=f.finished.strftime("%Y-%m-%dT%H:%M:%SZ") if f.finished else None,
                              errors=any((r.jobs_error > 0 for r in f.runs)))
                data['last_completed_flow'] = f_json

            f = self.ci_last_incomplete_flow
            if f:
                f_json = dict(id=f.id,
                              label=f.label,
                              created=f.created.strftime("%Y-%m-%dT%H:%M:%SZ"),
                              errors=any((r.jobs_error > 0 for r in f.runs)))
                data['last_incomplete_flow'] = f_json

        if with_cfg:
            data['stages'] = [s.get_json() for s in self.stages if s.deleted is None]
        return data

Index('ix_branches_name_project_id_not_deleted', Branch.name, Branch.project_id, postgresql_where=Branch.deleted.is_(None), unique=True)


# Sequence kinds:
# 0. KK_FLOW_SEQ
# 1. KK_CI_FLOW_SEQ
# 2. KK_DEV_FLOW_SEQ
# 3. KK_RUN_SEQ
# 4. KK_CI_RUN_SEQ
# 5. KK_DEV_RUN_SEQ
class BranchSequence(db.Model):
    __tablename__ = "branch_sequences"
    id = Column(Integer, primary_key=True)
    kind = Column(Integer, default=0)
    branch_id = Column(Integer, ForeignKey('branches.id'), nullable=False)
    branch = relationship('Branch', back_populates="sequences")
    stage_id = Column(Integer, ForeignKey('stages.id'), nullable=True)
    stage = relationship('Stage', back_populates="sequences")
    value = Column(Integer, default=0)

    def get_json(self):
        data = dict(id=self.id,
                    kind=self.kind,
                    branch_id=self.branch_id,
                    branch_name=self.branch.name,
                    stage_id=self.stage_id,
                    stage_name=self.stage.name if self.stage else None,
                    value=self.value)
        return data


class Secret(db.Model, DatesMixin):
    __tablename__ = "secrets"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)
    project = relationship('Project', back_populates="secrets")
    kind = Column(Integer, default=0)
    data = Column(JSONB)

    def get_json(self):
        data = dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    name=self.name,
                    project_id=self.project_id,
                    project_name=self.project.name,
                    kind=consts.SECRET_KINDS_NAME[self.kind])
        data.update(self.data)
        if 'key' in data:
            data['key'] = '******'
        if 'secret' in data:
            data['secret'] = '******'
        return data


# PLANNING

class Stage(db.Model, DatesMixin):
    __tablename__ = "stages"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    description = Column(UnicodeText)
    branch_id = Column(Integer, ForeignKey('branches.id'), nullable=False)
    branch = relationship('Branch', back_populates="stages")
    enabled = Column(Boolean, default=True)
    schema = Column(JSONB, nullable=False)
    schema_code = Column(UnicodeText)
    triggers = Column(JSONB)
    timeouts = Column(JSONB)
    repo_access_token = Column(UnicodeText)
    repo_branch = Column(UnicodeText)
    repo_url = Column(UnicodeText)
    schema_file = Column(UnicodeText)
    schema_from_repo_enabled = Column(Boolean, default=False)
    repo_error = Column(UnicodeText)
    repo_refresh_interval = Column(UnicodeText)
    repo_refresh_job_id = Column(UnicodeText)
    repo_state = Column(Integer, default=consts.REPO_STATE_OK)
    repo_version = Column(UnicodeText)
    git_clone_params = Column(UnicodeText)
    runs = relationship('Run', back_populates="stage")
    sequences = relationship("BranchSequence", back_populates="stage")
    # services

    def get_json(self, with_schema=True):
        data = dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    name=self.name,
                    description=self.description,
                    enabled=self.enabled,
                    schema_from_repo_enabled=self.schema_from_repo_enabled,
                    repo_url=self.repo_url,
                    repo_branch=self.repo_branch,
                    repo_access_token=self.repo_access_token,
                    repo_state=self.repo_state,
                    repo_error=self.repo_error,
                    repo_refresh_interval=self.repo_refresh_interval,
                    git_clone_params=self.git_clone_params,
                    repo_version=self.repo_version,
                    schema_file=self.schema_file)

        if with_schema:
            data['schema'] = self.schema
            data['schema_code'] = self.schema_code

        return data

    def __repr__(self):
        return "<Stage %s, '%s'>" % (self.id, self.name)

    def get_default_args(self):
        args = {}
        for param in self.schema['parameters']:
            args[param['name']] = param['default']
        return args

# class Config(db.Model, DatesMixin):
#     __tablename__ = "configs"
#     id = Column(Integer, primary_key=True)
#     name = Column(UnicodeText)
#     #vector =
#     environments = relationship("Environment", back_populates="config")
#     stage_id = Column(Integer, ForeignKey('stages.id'), nullable=False)
#     stage = relationship("Stage", back_populates="configs")

# class Environment(db.Model, DatesMixin):
#     __tablename__ = "environments"
#     id = Column(Integer, primary_key=True)
#     #system
#     #agents_group
#     config_id = Column(Integer, ForeignKey('configs.id'), nullable=False)
#     config = relationship("Config", back_populates="environments")
#     job_plan_id = Column(Integer, ForeignKey('job_plans.id'), nullable=False)
#     job_plan = relationship("JobPlan", back_populates="environments")

# EXECUTION



class RepoChanges(db.Model, DatesMixin):
    __tablename__ = "repo_changes"
    id = Column(Integer, primary_key=True)
    data = Column(JSONB, nullable=False)
    flow = relationship("Flow", back_populates="trigger_data")
    runs = relationship("Run", back_populates="repo_data")


def _get_report_entries(artifacts, infix):
    report_entries = []
    if (isinstance(artifacts, dict) and 'public' in artifacts and 'entries' in artifacts['public']):
        for rep_ent in artifacts['public']['entries']:
            url = '/bk/artifacts/public/%s/%s' % (infix, rep_ent)
            name = rep_ent.rsplit('/', 1)[-1].split('.', 1)[0].capitalize()
            rep = dict(name=name,
                       url=url)
            report_entries.append(rep)
    return report_entries


class Flow(db.Model, DatesMixin):
    __tablename__ = "flows"
    id = Column(Integer, primary_key=True)
    finished = Column(DateTime(timezone=True))
    state = Column(Integer, default=consts.FLOW_STATE_IN_PROGRESS)
    kind = Column(Integer, default=consts.FLOW_KIND_CI)  # 0 - CI, 1 - dev
    branch_name = Column(UnicodeText)
    branch_id = Column(Integer, ForeignKey('branches.id'), nullable=False)
    branch = relationship('Branch', foreign_keys=[branch_id])
    runs = relationship('Run', back_populates="flow", order_by="Run.created")
    args = Column(JSONB, nullable=False, default={})
    artifacts = Column(JSONB, default={})
    label = Column(UnicodeText)
    trigger_data_id = Column(Integer, ForeignKey('repo_changes.id'))
    trigger_data = relationship('RepoChanges')
    artifacts_files = relationship('Artifact', back_populates="flow")
    comments = relationship("TestCaseComment", back_populates="last_flow")
    summary = Column(JSONB, default={})

    Index('ix_flows_branch_id_kind', branch_id, kind)

    def get_label(self):
        return self.label if self.label else ("%d." % self.id)

    def get_json(self, with_project=True, with_branch=True, with_schema=True):
        if self.state == consts.FLOW_STATE_COMPLETED:
            duration = self.finished - self.created
        else:
            duration = utils.utcnow() - self.created

        trigger = None
        if self.trigger_data:
            trigger = self.trigger_data.data[0]

        data = dict(id=self.id,
                    label=self.get_label(),
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    finished=self.finished.strftime("%Y-%m-%dT%H:%M:%SZ") if self.finished else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    state=consts.FLOW_STATES_NAME[self.state],
                    kind='ci' if self.kind == 0 else 'dev',
                    duration=duration_to_txt(duration),
                    branch_name=self.branch_name,
                    args=self.args,
                    stages=[s.get_json(with_schema=with_schema) for s in self.branch.stages if s.deleted is None],
                    runs=[r.get_json(with_project=False, with_branch=False, with_artifacts=False) for r in self.runs],
                    trigger=trigger,
                    artifacts=self.artifacts)

        infix = 'f/%d' % self.id
        data['report_entries'] = _get_report_entries(self.artifacts, infix)

        if with_project:
            data['project_id'] = self.branch.project_id
            data['project_name'] = self.branch.project.name

        if with_branch:
            data['branch_id'] = self.branch_id
            data['base_branch_name'] = self.branch.name

        return data


Index('ix_flows_created', Flow.created)


class Run(db.Model, DatesMixin):
    __tablename__ = "runs"
    id = Column(Integer, primary_key=True)
    started = Column(DateTime(timezone=True))    # time when the run got a first non-deleted job
    finished = Column(DateTime(timezone=True))    # time when all jobs finished first time
    finished_again = Column(DateTime(timezone=True))    # time when all jobs finished
    state = Column(Integer, default=consts.RUN_STATE_IN_PROGRESS)
    email_sent = Column(DateTime(timezone=True))
    note = Column(UnicodeText)
    stage_id = Column(Integer, ForeignKey('stages.id'), nullable=False)
    stage = relationship('Stage', back_populates="runs")
    flow_id = Column(Integer, ForeignKey('flows.id'), nullable=False, index=True)
    flow = relationship('Flow', back_populates="runs")
    jobs = relationship('Job', back_populates="run")
    artifacts_files = relationship('Artifact', back_populates="run")
    hard_timeout_reached = Column(DateTime(timezone=True))
    soft_timeout_reached = Column(DateTime(timezone=True))
    args = Column(JSONB, nullable=False, default={})
    # stats - result changes
    fix_cnt = Column(Integer, default=0)
    new_cnt = Column(Integer, default=0)
    no_change_cnt = Column(Integer, default=0)
    regr_cnt = Column(Integer, default=0)
    issues_new = Column(Integer, default=0)
    issues_total = Column(Integer, default=0)
    jobs_error = Column(Integer, default=0)
    jobs_total = Column(Integer, default=0)
    tests_not_run = Column(Integer, default=0)
    tests_passed = Column(Integer, default=0)
    tests_total = Column(Integer, default=0)
    artifacts = Column(JSONB, default={})
    label = Column(UnicodeText)
    reason = Column(JSONB, nullable=False)
    repo_data_id = Column(Integer, ForeignKey('repo_changes.id'))
    repo_data = relationship('RepoChanges')
    processed_at = Column(DateTime(timezone=True))    # time when results analysis completed

    def get_json(self, with_project=True, with_branch=True, with_artifacts=True):
        jobs_processing = 0
        jobs_executing = 0
        jobs_waiting = 0
        jobs_error = 0
        jobs_total = 0

        if self.state == consts.RUN_STATE_PROCESSED:
            jobs_total = self.jobs_total
            jobs_error = self.jobs_error
            if self.started:
                begin = self.started
            else:
                begin = self.created
            if self.finished is None:
                log.error('PROBLEM WITH NONE TIMESTAMP in run %s', self)
                duration = utils.utcnow() - begin
            else:
                duration = self.finished - begin
        else:
            non_covered_jobs = Job.query.filter_by(run=self).filter_by(covered=False).all()
            jobs_total = len(non_covered_jobs)
            jobs_completed = 0
            last_time = None
            for job in non_covered_jobs:
                if job.state == consts.JOB_STATE_EXECUTING_FINISHED:
                    jobs_processing += 1
                elif job.state == consts.JOB_STATE_ASSIGNED:
                    jobs_executing += 1
                elif job.state != consts.JOB_STATE_COMPLETED:
                    jobs_waiting += 1
                elif job.state == consts.JOB_STATE_COMPLETED:
                    jobs_completed += 1
                    if last_time is None or (job.completed and job.completed > last_time):
                        last_time = job.completed

                if job.completion_status not in [consts.JOB_CMPLT_ALL_OK, None]:
                    jobs_error += 1

            if jobs_total == jobs_completed and last_time:
                duration = last_time - self.created
            else:
                duration = utils.utcnow() - self.created

        data = dict(id=self.id,
                    label=self.label,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    started=self.started.strftime("%Y-%m-%dT%H:%M:%SZ") if self.started else None,
                    finished=self.finished.strftime("%Y-%m-%dT%H:%M:%SZ") if self.finished else None,
                    processed_at=self.processed_at.strftime("%Y-%m-%dT%H:%M:%SZ") if self.processed_at else None,
                    duration=duration_to_txt(duration),
                    state=consts.RUN_STATES_NAME[self.state],
                    stage_name=self.stage.name,
                    stage_id=self.stage_id,
                    flow_id=self.flow_id,
                    flow_kind='ci' if self.flow.kind == 0 else 'dev',
                    flow_label=self.flow.get_label(),
                    args=self.args,
                    jobs_total=jobs_total,
                    jobs_waiting=jobs_waiting,
                    jobs_executing=jobs_executing,
                    jobs_processing=jobs_processing,
                    jobs_error=jobs_error,
                    tests_total=self.tests_total,
                    tests_passed=self.tests_passed,
                    tests_not_run=self.tests_not_run,
                    issues_total=self.issues_total,
                    issues_new=self.issues_new,
                    new_cnt=self.new_cnt,
                    no_change_cnt=self.no_change_cnt,
                    regr_cnt=self.regr_cnt,
                    fix_cnt=self.fix_cnt,
                    repo_data=self.repo_data.data if self.repo_data else None,
                    reason=self.reason['reason'],
                    note=self.note)

        if with_project:
            data['project_id'] = self.flow.branch.project_id
            data['project_name'] = self.flow.branch.project.name

        if with_branch:
            data['branch_id'] = self.flow.branch_id
            data['branch_name'] = self.flow.branch.name

        if with_artifacts:
            data['artifacts_total'] = len(self.artifacts_files)
            infix = 'r/%d' % self.id
            data['report_entries'] = _get_report_entries(self.artifacts, infix)

        return data


class Step(db.Model, DatesMixin):
    __tablename__ = "steps"
    id = Column(Integer, primary_key=True)
    index = Column(Integer, nullable=False)
    job_id = Column(Integer, ForeignKey('jobs.id'), nullable=False)
    job = relationship("Job", back_populates="steps")
    tool_id = Column(Integer, ForeignKey('tools.id'), nullable=False)
    tool = relationship("Tool", back_populates="steps")
    fields = Column(JSONB, nullable=False)
    result = Column(JSONB)
    status = Column(Integer)
    fields_masked = Column(JSONB, nullable=True)
    # services

    def get_json(self, mask_secrets=False):
        data = dict(id=self.id,
                    index=self.index,
                    tool=self.tool.name,
                    tool_id=self.tool_id,
                    tool_location=self.tool.location,
                    tool_entry=self.tool.entry,
                    tool_version=self.tool.version,
                    job_id=self.job_id,
                    status=self.status,
                    result=self.result)

        if mask_secrets and self.fields_masked:
            fields = self.fields_masked
        else:
            fields = self.fields

        name = self.tool.name
        if 'name' in fields:
            name = fields['name']
        elif self.tool.name == 'shell':
            if 'script' in fields and fields['script']:
                name = fields['script'].strip().splitlines()[0] + '...'
            elif 'cmd' in fields and fields['cmd']:
                name = fields['cmd']
        elif self.tool.name == 'git':
            name = 'checkout: ' + fields['checkout']

        data['name'] = name

        for f, v in fields.items():
            data[f] = v

        return data


class Job(db.Model, DatesMixin):
    __tablename__ = "jobs"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    assigned = Column(DateTime(timezone=True))
    started = Column(DateTime(timezone=True))
    finished = Column(DateTime(timezone=True))            # time when agent reported that job is finished
    processing_started = Column(DateTime(timezone=True))  # TODO: this is never used
    completed = Column(DateTime(timezone=True))
    run_id = Column(Integer, ForeignKey('runs.id'), nullable=False)
    run = relationship("Run", back_populates="jobs")
    steps = relationship("Step", back_populates="job", order_by="Step.index")
    state = Column(Integer, default=consts.JOB_STATE_QUEUED)
    completion_status = Column(Integer)
    covered = Column(Boolean, default=False)
    notes = Column(UnicodeText)
    agent = relationship('Agent', uselist=False, back_populates="job",
                            foreign_keys="Agent.job_id", post_update=True)
    agent_used_id = Column(Integer, ForeignKey('agents.id'))
    agent_used = relationship('Agent', foreign_keys=[agent_used_id], post_update=True)
    agents_group_id = Column(Integer, ForeignKey('agents_groups.id'), nullable=False)
    agents_group = relationship('AgentsGroup', back_populates="jobs")
    timeout = Column(Integer)
    system_id = Column(Integer, ForeignKey('systems.id', name='fk_systems_jobs'), nullable=False) # match name fk_systems_jobs with name in alembic migration
    system = relationship('System', back_populates="jobs")
    results = relationship('TestCaseResult', back_populates="job")
    issues = relationship('Issue', back_populates="job")

    def get_json(self, mask_secrets=False):
        if self.started:
            if self.finished:
                duration = self.finished - self.started
            else:
                duration = utils.utcnow() - self.started
            duration = duration_to_txt(duration)
        else:
            duration = ''

        steps = []
        for s in sorted(self.steps, key=lambda s: s.index):
            s = s.get_json(mask_secrets=mask_secrets)
            steps.append(s)

        return dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    started=self.started.strftime("%Y-%m-%dT%H:%M:%SZ") if self.started else None,
                    finished=self.finished.strftime("%Y-%m-%dT%H:%M:%SZ") if self.finished else None,
                    completed=self.completed.strftime("%Y-%m-%dT%H:%M:%SZ") if self.completed else None,
                    # processing_started=self.processing_started.strftime(
                    #     "%Y-%m-%dT%H:%M:%SZ") if self.processing_started else None,
                    duration=duration,
                    name=self.name,
                    state=self.state,
                    completion_status=self.completion_status,
                    timeout=self.timeout,
                    covered=self.covered,
                    notes=self.notes,
                    system_id=self.system_id,
                    system=self.system.name,
                    executor=self.system.executor,
                    run_id=self.run_id,
                    agents_group_id=self.agents_group_id,
                    agents_group_name=self.agents_group.name,
                    agent_id=self.agent_used_id if self.agent_used else 0,
                    agent_name=self.agent_used.name if self.agent_used else '',
                    steps=steps)

    def __repr__(self):
        txt = 'Job %s, state:%s' % (self.id, consts.JOB_STATES_NAME[self.state])
        txt += ', g:%s' % self.agents_group_id
        if self.agent_used_id:
            txt += ', ag:%s' % self.agent_used_id
        return "<%s>" % txt


class TestCase(db.Model, DatesMixin):
    __tablename__ = "test_cases"
    __test__ = False  # do not treat this class as a test by pytest
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText, unique=True)
    tool_id = Column(Integer, ForeignKey('tools.id'), nullable=False)
    tool = relationship('Tool', back_populates="test_cases")
    results = relationship("TestCaseResult", back_populates="test_case")
    comments = relationship("TestCaseComment", back_populates="test_case")


class TestCaseResult(db.Model):
    __tablename__ = "test_case_results"
    __test__ = False  # do not treat this class as a test by pytest
    id = Column(Integer, primary_key=True)
    test_case_id = Column(Integer, ForeignKey('test_cases.id'), nullable=False)
    test_case = relationship('TestCase', back_populates="results")
    job_id = Column(Integer, ForeignKey('jobs.id'), nullable=False)
    job = relationship('Job', back_populates="results")
    result = Column(Integer, default=0)
    values = Column(JSONB)
    cmd_line = Column(UnicodeText)
    instability = Column(Integer, default=0)
    age = Column(Integer, default=0)
    change = Column(Integer, default=consts.TC_RESULT_CHANGE_NO)
    relevancy = Column(Integer, default=0)
    comment_id = Column(Integer,
                        ForeignKey('test_case_comments.id',
                                   # this is WA, remember to set correct fk name in migration
                                   name='fk_test_case_comments_test_case_results'),
                        nullable=True)
    comment = relationship('TestCaseComment', back_populates="test_case_results")

    Index('ix_test_case_results_test_case_id', test_case_id)
    Index('ix_test_case_results_job_id', job_id)
    Index('ix_test_case_results_comment_id', comment_id)

    def __repr__(self):
        txt = 'TCR %s, result:%s' % (self.id, consts.TC_RESULTS_NAME[self.result])
        return "<%s>" % txt

    def get_json(self, with_extra=False, with_comment=False):
        data = dict(id=self.id,
                    test_case_id=self.test_case_id,
                    test_case_name=self.test_case.name,
                    result=self.result,
                    values=self.values,
                    cmd_line=self.cmd_line,
                    instability=self.instability,
                    age=self.age,
                    change=self.change,
                    relevancy=self.relevancy if self.relevancy is not None else 0,
                    job_id=self.job_id,
                    job_name=self.job.name,
                    agents_group_name=self.job.agents_group.name,
                    agents_group_id=self.job.agents_group_id,
                    system_name=self.job.system.name,
                    system_id=self.job.system_id,
                    agent_name=self.job.agent_used.name if self.job.agent_used else '',
                    agent_id=self.job.agent_used_id if self.job.agent_used else 0)

        if with_extra:
            data['project_id'] = self.job.run.flow.branch.project_id
            data['project_name'] = self.job.run.flow.branch.project.name
            data['branch_id'] = self.job.run.flow.branch_id
            data['branch_name'] = self.job.run.flow.branch.name
            data['flow_id'] = self.job.run.flow_id
            data['flow_kind'] = 'ci' if self.job.run.flow.kind == 0 else 'dev'
            data['flow_label'] = self.job.run.flow.get_label()
            created_at = self.job.run.flow.created
            data['flow_created_at'] = created_at.strftime("%Y-%m-%dT%H:%M:%SZ") if created_at else None
            data['run_id'] = self.job.run_id
            data['stage_id'] = self.job.run.stage_id
            data['stage_name'] = self.job.run.stage.name

        if with_comment and self.comment and self.comment.data:
            data['comment'] = dict(state=self.comment.state,
                                   data=list(reversed(self.comment.data)))

        return data


class TestCaseComment(db.Model):
    __tablename__ = "test_case_comments"
    id = Column(Integer, primary_key=True)
    test_case_id = Column(Integer, ForeignKey('test_cases.id'), nullable=True)
    test_case = relationship('TestCase', back_populates="comments")
    branch_id = Column(Integer, ForeignKey('branches.id'), nullable=True)
    branch = relationship('Branch', back_populates="comments")
    last_flow_id = Column(Integer, ForeignKey('flows.id'), nullable=True)
    last_flow = relationship('Flow', back_populates="comments")
    test_case_results = relationship('TestCaseResult', back_populates="comment")
    state = Column(Integer, default=consts.TC_COMMENT_NEW)
    data = Column(JSONB)

    def get_json(self):
        return dict(id=self.id,
                    state=self.state,
                    data=self.data)


class Issue(db.Model):
    __tablename__ = "issues"
    id = Column(Integer, primary_key=True)
    issue_type = Column(Integer, default=consts.ISSUE_TYPE_ERROR)
    line = Column(Integer)
    column = Column(Integer)
    path = Column(UnicodeText)
    symbol = Column(UnicodeText)
    message = Column(UnicodeText)
    job_id = Column(Integer, ForeignKey('jobs.id'), nullable=False)
    job = relationship('Job', back_populates="issues")
    extra = Column(JSONB)
    age = Column(Integer, default=0)

    def get_json(self):
        data = dict(id=self.id,
                    issue_type=self.issue_type,
                    line=self.line,
                    column=self.column,
                    path=self.path,
                    symbol=self.symbol,
                    message=self.message,
                    age=self.age,
                    job_id=self.job_id,
                    job_name=self.job.name,
                    agents_group_name=self.job.agents_group.name,
                    agents_group_id=self.job.agents_group_id,
                    agent_name=self.job.agent_used.name,
                    agent_id=self.job.agent_used_id)
        if self.extra:
            data.update(self.extra)
        return data


class File(db.Model):
    __tablename__ = "files"
    id = Column(Integer, primary_key=True)
    path = Column(UnicodeText)
    artifacts = relationship('Artifact', back_populates="file")


class Artifact(db.Model):
    __tablename__ = "artifacts"
    id = Column(Integer, primary_key=True)
    file_id = Column(Integer, ForeignKey('files.id'), nullable=False)
    file = relationship('File', back_populates="artifacts")
    flow_id = Column(Integer, ForeignKey('flows.id'), nullable=False, index=True)
    flow = relationship('Flow', back_populates="artifacts_files")
    run_id = Column(Integer, ForeignKey('runs.id'), nullable=False, index=True)
    run = relationship('Run', back_populates="artifacts_files")
    size = Column(Integer, default=0)
    section = Column(Integer, default=0)

    def get_json(self):
        return dict(id=self.id,
                    path=self.file.path,
                    size=self.size,
                    flow_id=self.flow_id,
                    run_id=self.run_id,
                    stage=self.run.stage.name)


class APSchedulerJob(db.Model):
    __tablename__ = 'apscheduler_jobs'
    id = Column(Unicode(191), autoincrement=False, nullable=False, primary_key=True)
    next_run_time = Column(DOUBLE_PRECISION(precision=53), autoincrement=False, nullable=True, index=True, unique=False)
    job_state = Column(BYTEA, autoincrement=False, nullable=False)


# RESOURCES

class System(db.Model):
    __tablename__ = "systems"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    executor = Column(UnicodeText)
    jobs = relationship('Job', back_populates="system")
    UniqueConstraint(name, executor, name='uq_system_name_executor')

    def get_json(self):
        return dict(id=self.id,
                    name=self.name,
                    executor=self.executor)


class Tool(db.Model, DatesMixin):
    __tablename__ = "tools"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    description = Column(UnicodeText)
    configuration = Column(UnicodeText)
    steps = relationship("Step", back_populates="tool")
    fields = Column(JSONB, nullable=False)
    # TODO should it have optional reference to project so that there are local and global tools?
    test_cases = relationship("TestCase", back_populates="tool")
    location = Column(UnicodeText)
    entry = Column(UnicodeText)
    version = Column(UnicodeText)
    url = Column(UnicodeText)
    tag = Column(UnicodeText)
    tool_file = Column(UnicodeText)
    UniqueConstraint(name, version, name='uq_tool_name_version')

    def __repr__(self):
        return "<Tool %s, '%s'>" % (self.id, self.name)

    def get_json(self, with_details=False):
        data = dict(id=self.id,
                    name=self.name)
        if with_details:
            data['location'] = self.location
            data['entry'] = self.entry
            data['version'] = self.version
            data['description'] = self.description
            data['fields'] = self.fields
            data['url'] = self.url
            data['tag'] = self.tag
            data['tool_file'] = self.tool_file
        return data


class AgentAssignment(db.Model):
    __tablename__ = "agent_assignments"
    agent_id = Column(Integer, ForeignKey('agents.id'), primary_key=True)
    agent = relationship('Agent', back_populates="agents_groups")
    agents_group_id = Column(Integer, ForeignKey('agents_groups.id'), primary_key=True)
    agents_group = relationship('AgentsGroup', back_populates="agents")


class AgentsGroup(db.Model, DatesMixin):
    __tablename__ = "agents_groups"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=True)
    project = relationship('Project', back_populates="agents_groups")
    # agents = relationship("Agent", back_populates="agents_group")  # static assignments
    agents = relationship('AgentAssignment', back_populates="agents_group")
    jobs = relationship("Job", back_populates="agents_group")
    deployment = Column(JSONB)
    extra_attrs = Column(JSONB)

    def get_json(self):
        deployment = self.deployment
        if not deployment:
            deployment = dict(method=0)

        if 'aws' not in deployment:
            deployment['aws'] = dict(region='', instances_limit=5, default_image='', instance_type='', disk_size=0,
                                     destruction_after_jobs=1, destruction_after_time=30)
        else:
            if 'destruction_after_jobs' not in deployment['aws']:
                deployment['aws']['destruction_after_jobs'] = 1
            if 'destruction_after_time' not in deployment['aws']:
                deployment['aws']['destruction_after_time'] = 30
            if 'disk_size' not in deployment['aws']:
                deployment['aws']['disk_size'] = 0

        if 'aws_ecs_fargate' not in deployment:
            deployment['aws_ecs_fargate'] = dict(region='', instances_limit=5, cluster='', subnets='', security_groups='')

        if 'azure_vm' not in deployment:
            deployment['azure_vm'] = dict(location='', instances_limit=5, default_image='', vm_size='',
                                          destruction_after_jobs=1, destruction_after_time=30)

        if 'kubernetes' not in deployment:
            deployment['kubernetes'] = dict(instances_limit=5)

        return dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    name=self.name,
                    project_id=self.project_id,
                    project_name=self.project.name if self.project else None,
                    agents_count=len([a for a in self.agents if not a.agent.deleted]),
                    deployment=deployment)

    def get_deployment(self):
        method = self.deployment['method']
        if method == consts.AGENT_DEPLOYMENT_METHOD_AWS_EC2:
            depl = self.deployment['aws']
        elif method == consts.AGENT_DEPLOYMENT_METHOD_AWS_ECS_FARGATE:
            depl = self.deployment['aws_ecs_fargate']
        elif method == consts.AGENT_DEPLOYMENT_METHOD_AZURE_VM:
            depl = self.deployment['azure_vm']
        elif method == consts.AGENT_DEPLOYMENT_METHOD_K8S:
            depl = self.deployment['kubernetes']
        else:
            msg = 'deployment method %d in agents group id:%d not implemented' % (self.deployment['method'], self.id)
            raise Exception(msg)
        return method, depl




class Agent(db.Model, DatesMixin):
    __tablename__ = "agents"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText, nullable=False)
    address = Column(UnicodeText, index=True, nullable=False, unique=True)
    ip_address = Column(UnicodeText)
    state = Column(Integer, default=0)
    disabled = Column(Boolean, default=False)
    comment = Column(UnicodeText)
    status_line = Column(UnicodeText)
    job_id = Column(Integer, ForeignKey('jobs.id'))
    job = relationship('Job', back_populates="agent", foreign_keys=[job_id])
    authorized = Column(Boolean, default=False)
    last_seen = Column(DateTime(timezone=True))
    host_info = Column(JSONB)
    user_attrs = Column(JSONB)
    extra_attrs = Column(JSONB)
    agents_groups = relationship('AgentAssignment', back_populates="agent")

    def __repr__(self):
        return "<Agent %s, job:%s>" % (self.id, self.job_id)

    def get_json(self):
        return dict(id=self.id,
                    created=self.created.strftime("%Y-%m-%dT%H:%M:%SZ") if self.created else None,
                    deleted=self.deleted.strftime("%Y-%m-%dT%H:%M:%SZ") if self.deleted else None,
                    last_seen=self.last_seen.strftime("%Y-%m-%dT%H:%M:%SZ") if self.last_seen else None,
                    name=self.name,
                    address=self.address,
                    authorized=self.authorized,
                    ip_address=self.ip_address,
                    state=self.state,
                    disabled=self.disabled,
                    comment=self.comment,
                    status_line=self.status_line,
                    host_info=self.host_info,
                    user_attrs=self.user_attrs,
                    extra_attrs=self.extra_attrs,
                    groups=[dict(id=a.agents_group.id, name=a.agents_group.name) for a in self.agents_groups],
                    job=self.job.get_json() if self.job else None)


class Setting(db.Model):
    __tablename__ = "settings"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    value = Column(UnicodeText)
    val_type = Column(UnicodeText)  # integer, text, boolean, password
    group = Column(UnicodeText)

    def get_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "value": self.get_value(),
            "type": self.val_type}

    def get_value(self, password_blank=True):
        if self.val_type == "integer":
            return int(self.value)
        if self.val_type == "boolean":
            return self.value == 'True'
        if self.val_type == "password" and password_blank:
            return ''
        if self.value is None:
            return ''
        return self.value

    def set_value(self, value):
        if self.val_type == "integer":
            self.value = str(value)
        elif self.val_type == "boolean":
            self.value = str(value)
        else:
            self.value = value


def get_setting(group, name):
    s = Setting.query.filter_by(group=group, name=name).one_or_none()
    if s is None:
        raise Exception('cannot find setting %s:%s' % (group, name))

    return s.get_value(password_blank=False)


def get_settings_group(group):
    ss = Setting.query.filter_by(group=group).all()
    if not ss:
        raise Exception('cannot find settings in %s group' % group)

    resp = {}
    for s in ss:
        resp[s.name] = s.get_value(password_blank=False)
    return resp


def set_setting(group, name, val):
    s = Setting.query.filter_by(group=group, name=name).one_or_none()
    if s is None:
        raise Exception('cannot find setting %s:%s' % (group, name))

    s.value = val
    db.session.commit()


class User(db.Model, DatesMixin):
    __tablename__ = "users"
    id = Column(Integer, primary_key=True)
    name = Column(UnicodeText)
    password = Column(UnicodeText)
    sessions = relationship("UserSession", back_populates="user")
    details = Column(JSONB)

    def get_json(self):
        if self.details:
            details = self.details
        else:
            details = {}

        return dict(id=self.id,
                    name=self.name,
                    enabled=details.get('enabled', True),
                    email=details.get('email', ''),
                    superadmin=self.name == 'admin')


class UserSession(db.Model, DatesMixin):
    __tablename__ = "user_sessions"
    id = Column(Integer, primary_key=True)
    token = Column(UnicodeText)
    user_id = Column(Integer, ForeignKey('users.id'), nullable=True)
    user = relationship('User', back_populates="sessions")
    details = Column(JSONB)

    def get_json(self):
        return dict(id=self.id,
                    token=self.token,
                    user=self.user.get_json() if self.user else None)


class CasbinRule(db.Model, DatesMixin):
    __tablename__ = "casbin_rules"

    id = Column(Integer, primary_key=True)
    ptype = Column(String(255))
    v0 = Column(String(255))
    v1 = Column(String(255))
    v2 = Column(String(255))
    v3 = Column(String(255))
    v4 = Column(String(255))
    v5 = Column(String(255))

    def __str__(self):
        arr = [self.ptype]
        for v in (self.v0, self.v1, self.v2, self.v3, self.v4, self.v5):
            if v is None:
                break
            arr.append(v)
        return ", ".join(arr)

    def __repr__(self):
        return '<CasbinRule {}: "{}">'.format(self.id, str(self))



# Copyright 2020 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import socket
from urllib.parse import urlparse

from . import utils


def _is_service_open(addr, port, sock_type):
    with socket.socket(socket.AF_INET, sock_type) as s:
        s.settimeout(1)
        try:
            s.connect((addr, port))
            s.shutdown(socket.SHUT_RDWR)
            return True
        except Exception:
            return False


def _parse_url_or_addr(url_addr, default_port=None):
    o = urlparse(url_addr)
    if o.hostname:
        host = o.hostname
        port = o.port or default_port
    else:
        host, port = utils.split_host_port(url_addr, default_port)

    return host, port


def is_service_open(url_addr, default_port=None):
    host, port = _parse_url_or_addr(url_addr, default_port)
    return _is_service_open(host, port, socket.SOCK_STREAM)


def wait_for_service(name, url_addr, default_port):
    host, port = _parse_url_or_addr(url_addr, default_port)

    attempt = 1
    trace = "checking TCP service %s on %s:%d..." % (name, host, port)
    print("%s %d." % (trace, attempt))
    while not _is_service_open(host, port, socket.SOCK_STREAM):
        if attempt < 3:
            time.sleep(2)
        elif attempt < 10:
            time.sleep(5)
        else:
            time.sleep(30)
        attempt += 1
        print("%s %d." % (trace, attempt))
    print("%s is up" % name)


def check_postgresql(url):
    wait_for_service('postgresql', url, 5432)



"""added details to user session

Revision ID: 9666bf8c1bb0
Revises: 2996b1b6e227
Create Date: 2022-10-23 08:20:15.282445

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9666bf8c1bb0'
down_revision = '2996b1b6e227'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('user_sessions', sa.Column('details', postgresql.JSONB(astext_type=sa.Text()), nullable=True))


def downgrade():
    op.drop_column('user_sessions', 'details')



# Copyright 2020-2022 The Kraken Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from flask import abort
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.orm.attributes import flag_modified

from . import consts
from . import utils
from .models import db, Branch, Flow, Run, Job, TestCaseResult
from .models import TestCase, TestCaseComment, System, AgentsGroup
from . import access

log = logging.getLogger(__name__)


def get_run_results(run_id, start=0, limit=10, sort_field="name", sort_dir="asc",
                    statuses=None, changes=None,
                    min_age=None, max_age=None,
                    min_instability=None, max_instability=None,
                    test_case_text=None, job=None,
                    systems=None, groups=None, token_info=None):

    run = Run.query.filter_by(id=run_id).one_or_none()
    if run is None:
        abort(404, "Run %s not found" % run_id)
    access.check(token_info, run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power user or project viewer roles can view results')

    log.info('filters %s %s %s %s %s %s %s %s %s %s',
             statuses, changes, min_age, max_age, min_instability, max_instability,
             test_case_text, job, systems, groups)
    q = TestCaseResult.query
    q = q.options(joinedload('test_case'),
                  joinedload('job'),
                  joinedload('job.agents_group'),
                  joinedload('job.agent_used'))
    q = q.join('job')
    q = q.filter(Job.run_id == run_id, Job.covered.is_(False))
    if statuses:
        q = q.filter(TestCaseResult.result.in_(statuses))
    if changes:
        q = q.filter(TestCaseResult.change.in_(changes))
    if min_age is not None:
        q = q.filter(TestCaseResult.age >= min_age)
    if max_age is not None:
        q = q.filter(TestCaseResult.age <= max_age)
    if min_instability is not None:
        q = q.filter(TestCaseResult.instability >= min_instability)
    if max_instability is not None:
        q = q.filter(TestCaseResult.instability <= max_instability)
    if test_case_text is not None:
        q = q.join('test_case').filter(TestCase.name.ilike('%' + test_case_text + '%'))
    if job is not None:
        if job.isdigit():
            job_id = int(job)
            q = q.filter(Job.id == job_id)
        else:
            q = q.filter(Job.name.ilike('%' + job + '%'))
    if systems:
        q = q.join('job', 'system').filter(System.id.in_(systems))
    if groups:
        q = q.join('job', 'agents_group').filter(AgentsGroup.id.in_(groups))

    total = q.count()

    sort_func = asc
    if sort_dir == "desc":
        sort_func = desc

    if sort_field == "result":
        q = q.order_by(sort_func('result'))
    elif sort_field == "change":
        q = q.order_by(sort_func('change'))
    elif sort_field == "age":
        q = q.order_by(sort_func('age'))
    elif sort_field == "instability":
        q = q.order_by(sort_func('instability'))
    elif sort_field == "relevancy":
        q = q.order_by(sort_func('relevancy'))
    elif sort_field == "system":
        q = q.join('job', 'system').order_by(sort_func('name'))
    elif sort_field == "group":
        q = q.join('job', 'agents_group').order_by(sort_func('relevancy'))
    else:
        q = q.join('test_case').order_by(sort_func('name'))

    q = q.offset(start).limit(limit)
    results = []
    for tcr in q.all():
        results.append(tcr.get_json(with_comment=True))
    return {'items': results, 'total': total}, 200


def get_result_history(test_case_result_id, start=0, limit=10, token_info=None):
    tcr = TestCaseResult.query.filter_by(id= test_case_result_id).one_or_none()
    if tcr is None:
        abort(404, "Test case result %s not found" % test_case_result_id)
    access.check(token_info, tcr.job.run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power user or project viewer roles can view results')

    q = TestCaseResult.query
    q = q.options(joinedload('test_case'),
                  joinedload('job'),
                  joinedload('job.agents_group'),
                  joinedload('job.agent_used'))
    q = q.filter_by(test_case_id=tcr.test_case_id)
    q = q.join('job')
    q = q.filter_by(agents_group=tcr.job.agents_group)
    q = q.filter_by(system=tcr.job.system)
    q = q.join('job', 'run', 'flow', 'branch')
    q = q.filter(Branch.id == tcr.job.run.flow.branch_id)
    q = q.filter(Flow.kind == 0)  # CI
    q = q.filter(Flow.created <= tcr.job.run.flow.created)
    q = q.order_by(desc(Flow.created))

    total = q.count()
    q = q.offset(start).limit(limit)
    results = []
    if tcr.job.run.flow.kind == 1:  # DEV
        results.append(tcr.get_json(with_extra=True))
    for tcr in q.all():
        results.append(tcr.get_json(with_extra=True))
    return {'items': results, 'total': total}, 200


def get_result(test_case_result_id, token_info=None):
    tcr = TestCaseResult.query.filter_by(id=test_case_result_id).one_or_none()
    if tcr is None:
        abort(404, "Test case result %s not found" % test_case_result_id)

    access.check(token_info, tcr.job.run.stage.branch.project_id, 'view',
                 'only superadmin, project admin, project power user or project viewer roles can view results')

    return tcr.get_json(with_extra=True), 200


def create_or_update_test_case_comment(test_case_result_id, body, token_info=None):
    author = body.get('author', None)
    state = body.get('state', None)
    text = body.get('text', None)
    if author is None:
        abort(400, "Missing author in request")
    if state is None:
        abort(400, "Missing state in request")
    if text is None:
        abort(400, "Missing text in request")

    tcr = TestCaseResult.query.filter_by(id=test_case_result_id).one_or_none()
    if tcr is None:
        abort(404, "Test case result %s not found" % test_case_result_id)

    access.check(token_info, tcr.job.run.stage.branch.project_id, 'pwrusr',
                 'only superadmin, project admin and project power user roles can comment results')

    if tcr.comment:
        tcc = tcr.comment
    else:
        # find if there exists a TestCaseComment for this test case and this branch
        q = TestCaseComment.query
        q = q.filter_by(test_case=tcr.test_case)
        q = q.filter_by(branch=tcr.job.run.flow.branch)
        tcc = q.one_or_none()

        if tcc is None:
            tcc = TestCaseComment(test_case=tcr.test_case,
                                  branch=tcr.job.run.flow.branch,
                                  last_flow=tcr.job.run.flow)
        elif tcc.last_flow.created < tcr.job.run.flow.created:
            tcc.last_flow = tcr.job.run.flow

        tcr.comment = tcc

    # find all other tcrs for the same test case and flow as current tcr
    # and assign this comment to them
    q = TestCaseResult.query
    q = q.filter_by(test_case=tcr.test_case)
    q = q.join('job', 'run', 'flow')
    q = q.filter(Flow.id == tcr.job.run.flow_id)
    for tcr2 in q.all():
        tcr2.comment = tcc

    # adjust relevancy is root cause found
    if (tcc.state not in [consts.TC_COMMENT_BUG_IN_PRODUCT, consts.TC_COMMENT_BUG_IN_TEST] and
        state in [consts.TC_COMMENT_BUG_IN_PRODUCT, consts.TC_COMMENT_BUG_IN_TEST]):
        tcr.relevancy -= 1

    # set user provided values in the comment
    tcc.state = state
    if not tcc.data:
        tcc.data = []
    tcc.data.append({
        'author': author,
        'date': utils.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        'text': text,
        'state': state,
        'tcr': test_case_result_id
    })
    flag_modified(tcc, 'data')
    db.session.commit()

    resp = tcc.get_json()
    resp['data'] = list(reversed(resp['data']))

    return resp, 200


def get_flow_analysis(flow_id, token_info=None):
    flow = Flow.query.filter_by(id=flow_id).one_or_none()
    if flow is None:
        abort(404, "Flow %s not found" % flow_id)
    access.check(token_info, flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user or project viewer roles can view results')

    q = Run.query.filter_by(flow_id=flow_id)

    recs_map = dict(systems={}, groups={})
    stats = {}
    analysis = dict(recs_map=recs_map,
                    stats=stats,
                    total_tests=0)

    total_tests = 0
    for run in q.all():
        log.info('run %s', run)

        if run.stage.name not in stats:
            stats[run.stage.name] = {'run_id': run.id}
        stage = stats[run.stage.name]

        q = TestCaseResult.query
        q = q.options(joinedload('test_case'),
                      joinedload('job'),
                      joinedload('job.agents_group'),
                      joinedload('job.agent_used'))
        q = q.join('job')
        q = q.filter(Job.run_id == run.id, Job.covered.is_(False))

        for tcr in q.all():
            # stage -> group -> system -> config -> component: PASSED: 5, NOT-RUN: 3, TOTAL: 8
            group_name = tcr.job.agents_group.name
            recs_map['groups'][group_name] = tcr.job.agents_group.id
            if group_name not in stage:
                stage[group_name] = {}
            group = stage[group_name]

            system_name = tcr.job.system.name
            recs_map['systems'][system_name] = tcr.job.system.id
            if system_name not in group:
                group[system_name] = {}
            system = group[system_name]

            # config_name = 'default'
            # component = 'all'

            if 'total' not in system:
                system['total'] = 0
            system['total'] += 1

            tcr_res = consts.TC_RESULTS_NAME[tcr.result]
            if tcr_res not in system:
                system[tcr_res] = 0
            system[tcr_res] += 1

            total_tests += 1

    analysis['total_tests'] = total_tests


    return analysis, 200


def get_branch_history(flow_id, limit=30, token_info=None):
    flow = Flow.query.filter_by(id=flow_id).one_or_none()
    if flow is None:
        abort(404, "Flow %s not found", flow_id)

    access.check(token_info, flow.branch.project_id, 'view',
                 'only superadmin, project admin, project power user or project viewer roles can view results')

    if flow.kind != consts.FLOW_KIND_CI:
        abort(400, "Branch history works only for CI, not DEV")

    if limit > 100:
        abort(400, "Cannot get history longer than 100 flows, requested %s", limit)

    q = Flow.query.filter(Flow.created <= flow.created)
    q = q.filter_by(branch_id=flow.branch_id)
    q = q.filter_by(kind=consts.FLOW_KIND_CI)
    q = q.order_by(desc(Flow.created))
    q = q.limit(limit)

    flows = []
    for f in reversed(q.all()):
        if f.summary:
            s = f.summary
        else:
            s = dict(id=f.id, label=f.get_label())

        s['created'] = f.created
        s['finished'] = f.finished

        flows.append(s)

    resp = dict(items=flows, total=len(flows))
    return resp, 200
